/*********************************************************************/
/*********************************************************************/
/**                                                                 **/
/**                    FONCTIONS ARITHMETIQUES                      **/
/**                       (deuxieme partie)                         **/
/**                                                                 **/
/**                      copyright Babe Cool                        **/
/**                                                                 **/
/*********************************************************************/
/*********************************************************************/
/* $Id: arith2.c,v 1.922 1997/04/06 02:52:40 karim Exp karim $ */

#include  "genpari.h"
GEN arith_proto(long f(GEN), GEN x, int do_error);
GEN garith_proto(GEN f(GEN), GEN x);
GEN garith_proto2gs(GEN f(GEN,long), GEN x, long y);

static GEN ellfacteur(GEN n1);

/***********************************************************************/
/**                                                                   **/
/**                          PRIME NUMBERS                            **/
/**                                                                   **/
/***********************************************************************/

GEN
prime(long n)
{
  byteptr p = diffptr;
  long c, prime = 0;

  while (n--)
  {
    c = *p++;
    if (c) prime += c; else err(primer1);
  }
  return stoi(prime);
}

GEN
primes(long n)
{
  GEN y,z;
  byteptr p = diffptr;
  long c, prime = 0;

  z = y = cgetg(n+1,t_VEC);
  while (n--)
  {
    c= *p++;
    if (c) prime+=c; else err(primer1);
    *++z = (long) stoi(prime);
  }
  return y;
}

byteptr
initprimes(long maxnum)
{
  register long k,size = (maxnum+257) >> 1;
  byteptr p = (byteptr) xmalloc(size+1);
  register byteptr q,r,s,fin=p+size;

  if (size > 218136500)
    err(talker,"impossible to have prestored primes > 436273009");
  memset(p, 0, size + 1);
  for (r=q=p,k=1; r<fin; )
  {
    do { r+=k; k+=2; r+=k; } while (*++q);
    for (s=r; s<fin; s+=k) *s=1;
  }
  r=p; *r++=2; *r++=1; /* 2 and 3 */
  for (s=q=r-1; ; s=q)
  {
    do q++; while (*q);
    if (q>=fin) break;
    *r++ = (q-s) << 1;
  }
  *r++=0;
  return (byteptr) xrealloc(p,r-p,size);
}

static void
cleanprimetab(void)
{
  long i,j;
  
  for(i=NUMPRTBELT; i >= 1; i--)
    if (primetab[i] == zero)
      for(j=i; j <= NUMPRTBELT; j++) primetab[j]=primetab[j+1];
  primetab[NUMPRTBELT+1]=un; /* Il est indispensable de garder la sentinelle a 1 */
}

/* primes is a single element or a row vector with "primes" to add to primetab.
 * If the "prime" shares a non-trivial factor with another element in primetab, 
 * one takes this into account.
 */
GEN
addprimes(GEN prime)
{
  long i,tx = typ(prime), av = avma;
  GEN p1,p2;

  if (tx == t_INT)
  {
    if (signe(prime) < 0) prime=absi(prime);
    if (cmpis(prime,1) <= 0) { avma=av; return primetab; }
    p1=cgetg(1,t_VEC);
    for (i=1; i <= NUMPRTBELT && primetab[i] != un; i++)
    {
      p2=mppgcd((GEN) primetab[i], prime);
      if (! gcmp1(p2))
      {
	if (cmpii(prime,p2)) p1=concat(p1,p2);
	p1=concat(p1,divii((GEN)primetab[i],p2));
	killbloc((GEN)primetab[i]); primetab[i]=zero;
      }
    }
    if (i == NUMPRTBELT+1 && lg(p1) == 1)
      err(talker,"extra primetable overflows");
    primetab[i] = lclone(prime);
    cleanprimetab();
    if (lg(p1) > 1) addprimes(p1);
    avma=av; return primetab;
  }
  if (! is_vec_t(tx)) err(typeer,"addprime");
  for (i=1; i < lg(prime); i++)
    addprimes((GEN) prime[i]);
  return primetab;
}

GEN
removeprimes(GEN prime)
{
  long i, tx = typ(prime), av = avma;

  if (tx == t_INT)
  {
    if (signe(prime) < 0) prime=absi(prime); 
    if (cmpis(prime,1) <= 0) { avma=av; return primetab; }
    for (i=1; i <= NUMPRTBELT; i++)
      if (! cmpii((GEN) primetab[i], prime))
      {
        killbloc((GEN)primetab[i]); primetab[i]=zero;
        cleanprimetab(); avma=av; return primetab;
      }
    err(talker,"prime is not in primetable in removeprimes");
  }
  if (! is_vec_t(tx)) err(typeer,"removeprimes");
  for (i=1; i < lg(prime); i++)
    removeprimes((GEN) prime[i]);
  return primetab;
}
  
/***********************************************************************/
/**                                                                   **/
/**                    BASIC ARITHMETIC FUNCTIONS                     **/
/**                                                                   **/
/***********************************************************************/

GEN
gmu(GEN n)
{
  return arith_proto(mu,n,1);
}

/* Attention, le parametre doit etre une variable. */
#define pseudoprime(p) millerrabin(p,5*lgefint(p))

long
mu(GEN n)
{
  byteptr d = diffptr;
  unsigned char c;
  GEN p,f;
  long av = avma, s = 1;

  if (typ(n) != t_INT) err(arither1);
  if (!signe(n)) err(arither2);
  if (is_pm1(n)) return 1;
  n=absi(n);
  p=cgeti(3); p[1]=evalsigne(1) | evallgefint(3); p[2]=0;
  while ((c = *d++))
  {
    p[2]+=c;
    if (!mpdivis(n,p,n)) continue;
    if (divise(n,p)) { avma=av; return 0; }
    s = -s;
    if (is_pm1(n)) break;
  }
  while (is_not_pm1(n))
  {
    f=icopy(n);
    while (! (cmpii(sqri(p),f)>0 || pseudoprime(f))) f=ellfacteur(f);
    diviiz(n,f,n); if (divise(n,f)) { avma=av; return 0; }
    s = -s;
  }
  avma=av; return s;
}

GEN
gphi(GEN n)
{
  return garith_proto(phi,n);
}

GEN
phi(GEN n)
{
  byteptr d = diffptr;
  unsigned char c;
  GEN f,p,m;
  long av1,av2;

  if (typ(n) != t_INT) err(arither1);
  if (!signe(n)) err(arither2);

  if (is_pm1(n)) return gun;
  m=cgeti(lgefint(n)); affsi(1,m);
  av1=avma; n=absi(n);
  p=cgeti(3); p[1]=evalsigne(1) | evallgefint(3); p[2]=0;
  av2=avma;
  while ((c = *d++))
  {
    p[2]+=c;
    if (!mpdivis(n,p,n)) continue;
    muliiz(m,subis(p,1),m);
    while (mpdivis(n,p,n)) muliiz(m,p,m);
    if (is_pm1(n)) break;
    avma=av2;
  }
  while (is_not_pm1(n))
  {
    f=icopy(n);
    while ( ! (cmpii(sqri(p),f) > 0 || pseudoprime(f)) ) f = ellfacteur(f);
    mpdivis(n,f,n); muliiz(m,subis(f,1),m);
    while (mpdivis(n,f,n)) muliiz(m,f,m);
    avma=av2;
  }
  avma=av1; return m;
}

GEN
auxdecomp(GEN n, long all)
{
  byteptr d=diffptr;
  GEN p,z,p1,p2;
  long nb=0,nbx=0,j,k,lim,av,sn;
  long pp[] = { evaltyp(t_INT)|evallg(4), evalsigne(1)|evallgefint(3),0,0 };

  if (typ(n) != t_INT) err(arither1);
  sn=signe(n); if (!sn) err(arither2);
  if (sn<0) { stoi(-1); stoi(1); nb++; }
  if (is_not_pm1(n))
  {
    n = gclone(n);  setsigne(n,1);
    lim = (all<=1)? VERYBIGINT: all;
    while (*d && pp[2] <= lim)
    {
      pp[2] += *d++;
      if (!mpdivis(n,pp,n)) continue;
      nb++; k=1; while (mpdivis(n,pp,n)) k++;
      icopy(pp); stoi(k);
      if (is_pm1(n)) break;
    }
/* NEW CODE P.M. & M. H.*/
    j=1;
    while (!gcmp1(n) && !gcmp1((GEN)primetab[j]))
    {
      if (mpdivis(n,(GEN) primetab[j],n))
      {
	nb++; k=1;
	while (mpdivis(n,(GEN) primetab[j],n)) k++;
	icopy((GEN) primetab[j]); stoi(k);
      }
      j++;
    }
/* UNTIL HERE */
    if (is_pm1(n)) killbloc(n);
    else
    {
      long step=1;
      GEN f;

      /* pp = square of biggest p tried so far */
      av=avma; affii(sqri(pp),pp); avma=av;
      do
      {
	while (carrecomplet(n,&f))
          { affii(f,n); step<<=1; avma=av; }
        f = n;
	if (all == 1)
          while (cmpii(pp,f) <= 0 && !pseudoprime(f))
	    { f = ellfacteur(f); nbx=1; }

        if (f == n) /* no more factors */
        { 
          icopy(f); k = step;
          killbloc(n); n=NULL;
        }
        else
        {
          f = gerepileupto(av,f); k=0;
          while (mpdivis(n,f,n)) k += step;
          if (is_pm1(n)) { killbloc(n); n=NULL; }
        }
        nb++; stoi(k); av=avma;
      }
      while (n);
    }
  }
  p = (GEN) avma; z=cgetg(3,t_MAT);
  p1=cgetg(nb+1,t_COL); z[1]=(long)p1;
  p2=cgetg(nb+1,t_COL); z[2]=(long)p2;
  for (j=nb; j; j--)
  {
    p2[j] = (long)p; p += lg(p);
    p1[j] = (long)p; p += lg(p);
  }
  if (DEBUGLEVEL>3 && nbx) { fprintferr("\n"); flusherr(); }
  if (nb)
  {
    GEN p1old,p2old;
    av=avma;
    p1old=dummycopy(p1);
    p2old=dummycopy(p2);
    p=sindexsort(p1);
    for (j=1;j<=nb; j++)
    {
      p1[j]=p1old[p[j]];
      p2[j]=p2old[p[j]];
    }
    avma=av;
  }
  return z;
}

GEN
decomp(GEN n)
{
  return auxdecomp(n,1);
}

GEN
smallfact(GEN n)
{
  return boundfact(n,0);
}


GEN
gboundfact(GEN n, long lim)
{
  return garith_proto2gs(boundfact,n,lim);
}

GEN
boundfact(GEN n, long lim)
{
  GEN p1,p2,p3,p4,p5,y;
  long av,tetpil;

  if (lim<=1) lim=0;
  switch(typ(n))
  {
    case t_INT:
      return auxdecomp(n,lim);
    case t_FRACN:
      av=avma; n=gred(n); /* fall through */
    case t_FRAC:
      if (typ(n)==t_FRAC) av=avma;
      p1=auxdecomp((GEN)n[1],lim);
      p2=auxdecomp((GEN)n[2],lim);
      p4=concat((GEN)p1[1],(GEN)p2[1]);
      p5=concat((GEN)p1[2],gneg((GEN)p2[2]));
      p3=indexsort(p4);

      tetpil=avma; y=cgetg(3,t_MAT);
      y[1]=(long)extract(p4,p3);
      y[2]=(long)extract(p5,p3);
      return gerepile(av,tetpil,y);
  }
  err(arither1);
  return NULL; /* not reached */
}

GEN
divisors(GEN n)
{
  GEN nbdiv = gun, t = auxdecomp(n,1), p1 = (GEN) t[1], p2 = (GEN) t[2];
  long tetpil,av=avma,i,j,start, l = lg(p1);
  GEN p,t1,t2,t3;

  start = (l>1 && signe(p1[1]) < 0) ? 2 : 1;
  for (i=start; i<l; i++) nbdiv = mulis(nbdiv,1+itos((GEN)p2[i]));
  p = t = cgetg(itos(nbdiv)+1,t_VEC);
  *++p=un;
  for (i=start; i<l; i++)
    for (t1=t,j=itos((GEN)p2[i]); j; j--,t1=t2)
    {
      t2=p;
      for (t3=t1; t3<t2; )
	*++p = lmulii((GEN)(*++t3),(GEN)p1[i]);
    }
  tetpil=avma;
  return gerepile(av,tetpil,sort(t));
}

GEN
core(GEN n)
{
  long av=avma,tetpil,i;
  GEN fa,p1,p2,res=gun;

  fa=auxdecomp(n,1); p1=(GEN)fa[1]; p2=(GEN)fa[2];
  for (i=1; i<lg(p1); i++)
    if (mpodd((GEN)p2[i])) res=mulii(res,(GEN)p1[i]);
  tetpil=avma; return gerepile(av,tetpil,icopy(res));
}

GEN
core2(GEN n)
{
  long av=avma,tetpil,i;

  GEN fa,p1,p2,p3,res=gun,res2=gun,y;

  fa=auxdecomp(n,1); p1=(GEN)fa[1]; p2=(GEN)fa[2];
  for (i=1; i<lg(p1); i++)
  {
    p3=(GEN)p2[i];
    if (mpodd(p3)) res=mulii(res,(GEN)p1[i]);
    if (!gcmp1(p3)) res2=mulii(res2,gpui((GEN)p1[i],shifti(p3,-1),0));
  }
  tetpil=avma; y=cgetg(3,t_VEC); y[1]=lcopy(res); y[2]=lcopy(res2);
  return gerepile(av,tetpil,y);
}

GEN
core0(GEN n,long flag)
{
  return (flag)? core2(n) : core(n);
}

GEN
coredisc(GEN n)
{
  long av=avma,tetpil,r;
  GEN p1=core(n);

  r=mod4(p1); if (signe(p1)<0) r=4-r;
  if (r==2 || r==3) { tetpil=avma; return gerepile(av,tetpil,shifti(p1,2)); }
  return p1;
}

GEN
coredisc2(GEN n)
{
  long av=avma,tetpil,r;
  GEN y,p1,p2=core2(n);

  p1=(GEN)p2[1]; r=mod4(p1); if (signe(p1)<0) r=4-r;
  if (r==2 || r==3)
  {
    tetpil=avma; y=cgetg(3,t_VEC); 
    y[1]=lshifti(p1,2); y[2]=lmul2n((GEN)p2[2],-1);
    return gerepile(av,tetpil,y);
  }
  return p2;
}

GEN
coredisc0(GEN n,long flag)
{
  return (flag)? coredisc2(n) : coredisc(n);
}

GEN
gomega(GEN n)
{
  return arith_proto(omega,n,1);
}

long
omega(GEN n)
{
  byteptr d=diffptr;
  GEN p,f;
  long nb=0,av=avma,av2;

  if (typ(n) != t_INT) err(arither1);
  if (!signe(n)) err(arither2);
  if (is_pm1(n)) return 0;

  n=absi(n); p=cgeti(3);
  p[1]=evalsigne(1) | evallgefint(3); p[2]=0;
  av2=avma;
  while (*d)
  {
    p[2] += *d++;
    if (!mpdivis(n,p,n)) continue;
    nb++; while (mpdivis(n,p,n)); /* empty */
    if (is_pm1(n)) break;
  }
  while (is_not_pm1(n))
  {
    f=icopy(n);
    while (!(cmpii(sqri(p),f)>0 || pseudoprime(f))) f=ellfacteur(f);
    nb++; while (mpdivis(n,f,n)); /* empty */
    avma=av2;
  }
  avma=av; return nb;
}


GEN
gbigomega(GEN n)
{
  return arith_proto(bigomega,n,1);
}

long
bigomega(GEN n)
{
  byteptr d=diffptr;
  unsigned char c;
  GEN p,f;
  long nb=0,av=avma,av2;

  if (typ(n) != t_INT) err(arither1);
  if (!signe(n)) err(arither2);
  if (is_pm1(n)) return 0;

  n=absi(n); p=cgeti(3);
  p[1]=evalsigne(1)+evallgefint(3); p[2]=0;
  av2=avma;
  while ((c = *d++))
  {
    p[2]+=c; if (!mpdivis(n,p,n)) continue;
    do nb++; while (mpdivis(n,p,n));
    if (is_pm1(n)) break;
  }
  while (is_not_pm1(n))
  {
    f=icopy(n);
    while (!(cmpii(sqri(p),f)>0 || pseudoprime(f))) f = ellfacteur(f);
    while (mpdivis(n,f,n)) nb++;
    avma=av2;
  }
  avma=av; return nb;
}

GEN
gnumbdiv(GEN n)
{
  return garith_proto(numbdiv,n);
}

GEN
numbdiv(GEN n)
{
  byteptr d=diffptr;
  unsigned char c;
  GEN p,f,m,m1;
  long l,av=avma,av2,av3;

  if (typ(n) != t_INT) err(arither1);
  if (!signe(n)) err(arither2);
  if (is_pm1(n)) return gun;

  n=absi(n); p=cgeti(3);
  p[1]=evalsigne(1)+evallgefint(3); p[2]=0;
  av2=avma; m=stoi(1);
  while ((c = *d++))
  {
    p[2]+=c;
    l=1; while (mpdivis(n,p,n)) l++;
    av3=avma; m1=mulsi(l,m);
    if (lgefint (m1) == lgefint(m)) affii(m1,m);
    else m=gerepile(av2,av3,m1);
    if (is_pm1(n)) break;
  }
  while (is_not_pm1(n))
  {
    f=icopy(n);
    while (!(cmpii(sqri(p),f)>0 || pseudoprime(f))) f = ellfacteur(f);
    l=1; while (mpdivis(n,p,n)) l++;
    av3=avma; m=gerepile(av2,av3,mulsi(l,m));
  }
  return gerepile(av,av2,m);
}

GEN
gsumdiv(GEN n)
{
  return garith_proto(sumdiv,n);
}

GEN
sumdiv(GEN n)
{
  byteptr d=diffptr;
  unsigned char c;
  GEN p,f,m,m1;
  long av1=avma,av2,av3;

  if (typ(n) != t_INT) err(arither1);
  if (!signe(n)) err(arither2);
  if (is_pm1(n)) return gun;

  n=absi(n); p=cgeti(3);
  p[1]=evalsigne(1)+evallgefint(3); p[2]=0;
  av2=avma; m=gun;
  while ((c = *d++))
  {
    p[2]+=c; if (!mpdivis(n,p,n)) continue;
    m1=addsi(1,p);
    while (mpdivis(n,p,n)) m1=addsi(1,mulii(p,m1));
    av3=avma; m=gerepile(av2,av3,mulii(m1,m));
    if (is_pm1(n)) break;
  }
  while (is_not_pm1(n))
  {
    f=icopy(n);
    while (!(cmpii(sqri(p),f)>0 || pseudoprime(f))) f=ellfacteur(f);
    m1=gun;
    while (mpdivis(n,f,n)) m1=addsi(1,mulii(f,m1));
    av3=avma; m=gerepile(av2,av3,mulii(m1,m));
  }
  return gerepile(av1,av2,m);
}

GEN
gsumdivk(GEN n, long k)
{
  return garith_proto2gs(sumdivk,n,k);
}

GEN
sumdivk(GEN n, long k)
{
  byteptr d=diffptr;
  unsigned char c;
  GEN p,p1,f,m,m1,pk;
  long av1=avma,av2,av3,k1;

  if (typ(n) != t_INT) err(arither1);
  if (!signe(n)) err(arither2);
  if (is_pm1(n)) return gun;

  n=absi(n); k1=k; if (k<0) { k= -k; p1=gpuigs(n,k1); }
  p=cgeti(3); p[1] = evalsigne(1) | evallgefint(3); p[2]=0;
  av2=avma; m=gun;
  while ((c = *d++))
  {
    p[2]+=c; if (!mpdivis(n,p,n)) continue;
    pk=gpuigs(p,k); m1=addsi(1,pk);
    while (mpdivis(n,p,n)) m1=addsi(1,mulii(pk,m1));
    av3=avma; m=gerepile(av2,av3,mulii(m1,m));
    if (is_pm1(n)) break;
  }
  while (is_not_pm1(n))
  {
    f=icopy(n);
    while (!(cmpii(sqri(p),f)>0 || pseudoprime(f))) f=ellfacteur(f);
    pk=gpuigs(f,k); m1=gun;
    while (mpdivis(n,f,n)) m1=addsi(1,mulii(pk,m1));
    av3=avma; m=gerepile(av2,av3,mulii(m1,m));
  }
  av3=avma; return (k1>=0) ? gerepile(av1,av2,m) : gerepile(av1,av3,gmul(p1,m));
}


/***********************************************************************/
/**                                                                   **/
/**                        FACTORIZATION (ECM)                        **/
/**                                                                   **/
/***********************************************************************/

/* decomposition de nombres par la methode des courbes elliptiques.
 * La seule fonction exportee est ellfacteur.
 * Cette fonction renvoie un facteur non trivial de n.
 * On suppose en entree que n n'est pas premier, et n'est divisible par
 * aucun petit nombre premier (pas de facteur<65536,en tout cas.)
 */
static GEN n, gl, *W;
#define nbc 10

static int sisprime(long p)
{
  byteptr d = diffptr;
  long q = *d++;

  while (q <= (p/q)) if (p%q == 0) return 0; else q += *d++;
  return 1;
}

static int
elladd(GEN *X1, GEN *X2)
{
  GEN *Y1 = X1 + nbc, *Y2 = X2 + nbc, *A = W + nbc + 1;
  long i, av=avma;
  for (i = 0; i < nbc; i++)
    {
      subiiz(X1[i], X2[i], A[i]); if (signe(A[i])<0) addiiz(A[i],n,A[i]);
      modiiz(mulii(A[i], W[i]), n, W[i + 1]);
      avma=av;
    }
  if (!inversemodulo(W[nbc], n, &gl)) return 1;
  while (i--)
  {
    GEN lambda = modii(mulii(subii(Y1[i], Y2[i]), mulii(gl, W[i])), n);
    modiiz(subii(sqri(lambda), addii(X2[i], X1[i])), n, X2[i]);
    modiiz(subii(mulii(lambda, subii(X1[i], X2[i])), Y1[i]), n, Y2[i]);
    modiiz(mulii(gl, A[i]), n, gl);
  }
  avma = av; return 0;
}

static int
elldouble(GEN *X)
{
  GEN *Y = X + nbc, v, lambda;
  long i,av=avma;
  for (i = 0; i < nbc; i++)
    {
      modiiz(mulii(Y[i], W[i]), n, W[i+1]);
      avma = av;
    }
  if (!inversemodulo(W[nbc], n, &gl)) return 1;
  while (i--)
  {
    lambda = modii(mulii(addsi(1, mulsi(3, mulii(X[i], X[i]))),
			 mulii(gl,W[i])), n);
    lambda = shifti(mpodd(lambda) ? addii(lambda, n) : lambda, -1);
    v = modii(subii(sqri(lambda), shifti(X[i], 1)), n);
    modiiz(mulii(gl, Y[i]), n, gl);
    modiiz(subii(mulii(lambda, subii(X[i], v)), Y[i]), n, Y[i]);
    affii(v, X[i]);
  }
  avma = av; return 0;
}

static int
ellmult(long k, GEN *X, GEN *XAUX)
{
  if (k == 1)
    { for (k = 2 * nbc; k--; ) affii (X[k], XAUX[k]); return 0; }
  if (ellmult(k / 2, X, XAUX) || elldouble(X)) return 1;
  return k&1 ? elladd(XAUX, X) : 0;
}

static GEN
ellfacteur(GEN n1)
{
  static long TB1[] =
    { 156,405,962,2240,4778,9004,18437,34155,66596,133297,
       247988,374990,649996,1170924,1967442,3276490,5249667 };
  long B1, B2, dsn, dsnmax, a = 1, i, j, p, m, av, tf=lgefint(n1);
  GEN *X, *XD, *XDD, res = cgeti(tf);
  byteptr d;

  av=avma; n=absi(n1);
  W = 1 + (GEN *)cgetg(260 * nbc + 2, 17);
  W[0]=gun;
  for (i = 260 * nbc; i > 0; i--) W[i] = cgeti(tf);
  X = W + 2 * nbc + 1; XD = X + 2 * nbc; XDD = XD + 2 * nbc;
  dsn = (long)sqrt(9.6*(tf-2)); dsnmax = (long)(4.8 * (tf - 2));
  if (dsnmax > 41) dsnmax = 41;
  if (dsn < 9) dsn = 9; else if (dsn > dsnmax) dsn = dsnmax;

  for(;;)
  {
    for (i = 2 * nbc; i--; ) affsi(a++, X[i]);
    B1 = TB1[dsn/2 - 4]; B2 = 50 * B1;
    if (DEBUGLEVEL > 3)
      { fprintferr("dsn = %d, B1 = %d, B2 = %d\n", dsn, B1, B2); flusherr(); }
    for (d = diffptr, p = *d++; p < B1; p += *d++)
    {
      for (m = B2; m > p; m /= p)
	if (ellmult(p, X, XD) && cmpii(gl, n)) goto fin;
      if (*d == 0) break;
    }
    while (p < B1)
    {
      do p += 2; while (!(sisprime(p)));
      for (m = B2; m > p; m /= p)
	if (ellmult(p, X, XD) && cmpii(gl, n)) goto fin;
    }
    for (i = 0; i < 2*nbc; i++) affii (X[i], XD[i]);
    if (elldouble(XD) && cmpii(gl, n)) goto fin;
    if (ellmult(p, X, XDD) && cmpii(gl, n)) goto fin;
    for (i = 0; i < 2 * nbc; i++) affii (XD[i], XDD[i]);
    if (elldouble(XDD) && cmpii(gl, n)) goto fin;
    for (i = 4; i < 256; i += 2)
    {
      for (j = (i - 2) * nbc; j < i * nbc; j++) affii (XD[j], XDD[j]);
      if (elladd(XD, XD + i * nbc) && cmpii(gl, n)) goto fin;
    }
    while (p < B2)
    {
      if (*d == 0) break;
      p += *d;
      if (elladd(X + nbc * *d++, X) && cmpii(gl, n)) goto fin;
    }
    while (p < B2)
    {
      long oldp = p;
      do p += 2; while (!sisprime(p));
      if (elladd(X + (p - oldp) * nbc, X) && cmpii(gl, n)) goto fin;
    }
    if (dsn < dsnmax) dsn += 2;
 }
 fin:
  affii(cmpii(sqri(gl), n) > 0 ? divii(n,gl) : gl, res);
  if (DEBUGLEVEL > 3)
    { fprintferr("p = %d, facteur = ", p); outerr(res); flusherr(); }
  avma=av; return res;
}

/***********************************************************************/
/**                                                                   **/
/**                      BINARY QUADRATIC FORMS                       **/
/**                                                                   **/
/***********************************************************************/

GEN
qfi(GEN x, GEN y, GEN z)
{
  GEN t;

  if (typ(x) != t_INT || typ(y) != t_INT || typ(z) != t_INT) err(qfer1);
  t=cgetg(4,t_QFI); t[1]=lcopy(x); t[2]=lcopy(y); t[3]=lcopy(z);
  return t;
}


GEN
qfr(GEN x, GEN y, GEN z, GEN d)
{
  GEN t;

  if (typ(x) != t_INT || typ(y) != t_INT || typ(z) != t_INT ||
      typ(d) != t_REAL) err(qfer1);
  t=cgetg(5,t_QFR); t[1]=lcopy(x); t[2]=lcopy(y); t[3]=lcopy(z); t[4]=lcopy(d);
  return t;
}

GEN
Qfb0(GEN x, GEN y, GEN z, GEN d, long prec)
{
  long av=avma,s;
  GEN t;

  if (typ(x) != t_INT || typ(y) != t_INT || typ(z) != t_INT) err(qfer1);
  s=signe(subii(sqri(y),shifti(mulii(x,z),2))); avma=av;
  switch(s)
  {
    case -1:
      t=cgetg(4,t_QFI); t[1]=lcopy(x); t[2]=lcopy(y); t[3]=lcopy(z); break;
    case  0:
      err(talker,"zero discriminant in Qfb"); break;
    case  1:
      if (typ(d) != t_REAL)
	{ t=d; gaffect(t,d=cgetr(prec)); }
      t=cgetg(5,t_QFR);
      t[1]=lcopy(x); t[2]=lcopy(y); t[3]=lcopy(z); t[4]=lcopy(d);
  }
  return t;
}

GEN
compimag(GEN x, GEN y)
{
  long av,tetpil;
  GEN s,n,d,d1,d2,x1,x2,y1,y2,v1,v2,v3,b3,c3,m,z,p1,r;

  if (cmpii((GEN)x[1],(GEN)y[1])>0) { s=x; x=y; y=s; }
  av=avma; s=shifti(addii((GEN)x[2],(GEN)y[2]),-1); n=subii((GEN)y[2],s);
  d=bezout((GEN)y[1],(GEN)x[1],&y1,&x1); d1=bezout(s,d,&x2,&y2);
  v1=divii((GEN)x[1],d1); v2=divii((GEN)y[1],d1);
  d2=(gcmp1(d1))?gun:mppgcd(d1,mppgcd((GEN)x[3],mppgcd((GEN)y[3],n)));
  v3=mulii(d2,v1);
  m=addii(mulii(mulii(y1,y2),n),mulii((GEN)y[3],x2)); setsigne(m,-signe(m));
  r=modii(m,v3); b3=shifti((p1=mulii(v2,r)),1);
  c3=addii(mulii((GEN)y[3],d1),mulii(r,addii((GEN)y[2],p1)));
  z=cgetg(4,t_QFI); z[1]=lmulii(v3,v2);
  z[2]=laddii((GEN)y[2],b3); z[3]=ldivii(c3,v3);
  tetpil=avma; return gerepile(av,tetpil,redimag(z));
}

GEN
compreal(GEN x, GEN y)
{
  long av,tetpil;
  GEN s,n,d,d1,d2,x1,x2,y1,y2,v1,v2,v3,b3,c3,m,z,p1,r;

  av=avma; s=shifti(addii((GEN)x[2],(GEN)y[2]),-1); n=subii((GEN)y[2],s);
  d=bezout((GEN)y[1],(GEN)x[1],&y1,&x1); d1=bezout(s,d,&x2,&y2);
  v1=divii((GEN)x[1],d1); v2=divii((GEN)y[1],d1);
  d2=(gcmp1(d1))?gun:mppgcd(d1,mppgcd((GEN)x[3],mppgcd((GEN)y[3],n)));
  v3=mulii(d2,v1);
  m=addii(mulii(mulii(y1,y2),n),mulii((GEN)y[3],x2)); setsigne(m,-signe(m));
  r=modii(m,v3); b3=shifti((p1=mulii(v2,r)),1);
  c3=addii(mulii((GEN)y[3],d1),mulii(r,addii((GEN)y[2],p1)));
  z=cgetg(5,t_QFR); z[1]=lmulii(v3,v2);
  z[2]=laddii((GEN)y[2],b3); z[3]=ldivii(c3,v3);
  z[4]=laddrr((GEN)x[4],(GEN)y[4]); tetpil=avma;
  return gerepile(av,tetpil,redreal(z));
}

GEN
comprealraw(GEN x, GEN y)
{
  long av,tetpil;
  GEN s,n,d,d1,d2,x1,x2,y1,y2,v1,v2,v3,b3,c3,m,z,p1,r;

  av=avma; s=shifti(addii((GEN)x[2],(GEN)y[2]),-1); n=subii((GEN)y[2],s);
  d=bezout((GEN)y[1],(GEN)x[1],&y1,&x1); d1=bezout(s,d,&x2,&y2);
  v1=divii((GEN)x[1],d1); v2=divii((GEN)y[1],d1);
  d2=(gcmp1(d1))?gun:mppgcd(d1,mppgcd((GEN)x[3],mppgcd((GEN)y[3],n)));
  v3=mulii(d2,v1);
  m=addii(mulii(mulii(y1,y2),n),mulii((GEN)y[3],x2)); setsigne(m,-signe(m));
  r=modii(m,v3); b3=shifti((p1=mulii(v2,r)),1);
  c3=addii(mulii((GEN)y[3],d1),mulii(r,addii((GEN)y[2],p1)));
  z=cgetg(5,t_QFR); z[1]=lmulii(v3,v2);
  z[2]=laddii((GEN)y[2],b3); z[3]=ldivii(c3,v3);
  z[4]=laddrr((GEN)x[4],(GEN)y[4]); tetpil=avma;
  return gerepile(av,tetpil,gcopy(z));
}

/* l=floor((|d|/4)^(1/4)) */
GEN
nucomp(GEN x, GEN y, GEN l)
{
  long av=avma,tetpil,cz;
  GEN s,n,d,d1,v,u1,u,v1,v2,z,p1,p2,p3;
  GEN a,b,e,f,g,a1,a2,a3,b3,v3,q,t2,t3,c3,q1,q2,q3,q4;

  if (typ(x) != t_QFI || typ(y) != t_QFI) err(nucomper1);
  if (x==y) return nudupl(x,l);

  if (cmpii((GEN)x[1],(GEN)y[1])<0) { s=x; x=y; y=s; }
  s=shifti(addii((GEN)x[2],(GEN)y[2]),-1); n=subii((GEN)y[2],s);
  a1=(GEN)x[1]; a2=(GEN)y[1]; d=bezout(a2,a1,&u,&v);
  if (gcmp1(d)) { a=negi(gmul(u,n)); d1=d; }
  else
    if (divise(s,d))
    {
      a=negi(gmul(u,n)); d1=d; a1=divii(a1,d1);
      a2=divii(a2,d1); s=divii(s,d1);
    }
    else
    {
      d1=bezout(s,d,&u1,&v1);
      if (!gcmp1(d1))
      {
	a1=divii(a1,d1); a2=divii(a2,d1);
	s=divii(s,d1); d=divii(d,d1);
      }
      p1=resii((GEN)x[3],d); p2=resii((GEN)y[3],d);
      p3=modii(negi(mulii(u1,addii(mulii(u,p1),mulii(v,p2)))),d);
      a=subii(mulii(p3,divii(a1,d)),mulii(u,divii(n,d)));
    }
  a=modii(a,a1); p1=subii(a1,a); if (cmpii(a,p1)>0) a=negi(p1);
  v=gzero; d=a1; v2=gun; v3=a; cz=0;
  while (cmpii(absi(v3),l)>0)
  {
    q=dvmdii(d,v3,&t3); t2=subii(v,mulii(q,v2));
    v=v2; d=v3; v2=t2; v3=t3; cz++;
  }
  if (!cz)
  {
    q1=mulii(a2,v3); q2=addii(q1,n); f=divii(q2,d);
    g=divii(addii(mulii(v3,s),(GEN)y[3]),d);
    a3=mulii(d,a2); c3=addii(mulii(v3,f),mulii(g,d1));
    b3=addii(shifti(q1,1),(GEN)y[2]);
  }
  else
  {
    if (cz&1) { v3=negi(v3); v2=negi(v2); }
    b=divii(addii(mulii(a2,d),mulii(n,v)),a1); q1=mulii(b,v3);
    q2=addii(q1,n); f=divii(q2,d);
    e=divii(addii(mulii(s,d),mulii((GEN)y[3],v)),a1);
    q3=mulii(e,v2); q4=subii(q3,s); g=divii(q4,v);
    if (!gcmp1(d1)) { v2=mulii(d1,v2); v=mulii(d1,v); }
    a3=addii(mulii(d,b),mulii(e,v)); c3=addii(mulii(v3,f),mulii(g,v2));
    b3=addii(addii(q1,q2),mulii(d1,addii(q3,q4)));
  }
  z=cgetg(4,t_QFI); z[1]=(long)a3; z[2]=(long)b3; z[3]=(long)c3;
  tetpil=avma; return gerepile(av,tetpil,redimag(z));
}

/* l=floor((|d|/4)^(1/4)) */
GEN
nudupl(GEN x, GEN l)
{
  long av=avma,tetpil,cz;
  GEN u,v,d,d1,p1,a,b,c,b2,z,v2,v3,q,t2,t3,e,g;

  d1=bezout((GEN)x[2],(GEN)x[1],&u,&v);
  a=divii((GEN)x[1],d1); b=divii((GEN)x[2],d1);
  c=modii(negi(mulii(u,(GEN)x[3])),a); p1=subii(a,c);
  if (cmpii(c,p1)>0) c=negi(p1);
  v=gzero; d=a; v2=gun; v3=c; cz=0;
  while (cmpii(absi(v3),l)>0)
  {
    q=dvmdii(d,v3,&t3); t2=subii(v,mulii(q,v2));
    v=v2; d=v3; v2=t2; v3=t3; cz++;
  }
  if (!cz)
  {
    g=divii(addii(mulii(b,v3),(GEN)x[3]),d);
    z=cgetg(4,t_QFI); z[1]=(long)sqri(d);
    z[2]=laddii((GEN)x[2],shifti(mulii(d,v3),1));
    z[3]=laddii(sqri(v3),mulii(g,d1));
  }
  else
  {
    if (cz&1) { v=negi(v); d=negi(d); }
    e=divii(addii(mulii((GEN)x[3],v),mulii(b,d)),a);
    g=divii(subii(mulii(e,v2),b),v); b2=addii(mulii(e,v2),mulii(v,g));
    if (!gcmp1(d1)) { b2=mulii(d1,b2); v=mulii(d1,v); v2=mulii(d1,v2); }
    z=cgetg(4,t_QFI); z[1]=laddii(sqri(d),mulii(e,v));
    z[2]=laddii(b2,shifti(mulii(d,v3),1));
    z[3]=laddii(sqri(v3),mulii(g,v2));
  }
  tetpil=avma; return gerepile(av,tetpil,redimag(z));
}

GEN
sqcomp(GEN x)
{
  long av,tetpil;
  GEN d1,d2,x2,y2,v1,v3,b3,c3,m,z,p1,r;

  av=avma;
  d1=bezout((GEN)x[2],(GEN)x[1],&x2,&y2); v1=divii((GEN)x[1],d1);
  d2=(gcmp1(d1))?gun:mppgcd(d1,(GEN)x[3]); v3=mulii(d2,v1);
  m=mulii((GEN)x[3],x2); setsigne(m,-signe(m));
  r=modii(m,v3); b3=shifti((p1=mulii(v1,r)),1);
  c3=addii(mulii((GEN)x[3],d1),mulii(r,addii((GEN)x[2],p1)));
  z=cgetg(4,t_QFI); z[1]=lmulii(v3,v1);
  z[2]=laddii((GEN)x[2],b3); z[3]=ldivii(c3,v3);
  tetpil=avma; return gerepile(av,tetpil,redimag(z));
}

GEN
sqcompreal0(GEN x, long raw)
{
  long av,tetpil;
  GEN d1,d2,x2,y2,v1,v3,b3,c3,m,z,p1,r;

  if (typ(x) != t_QFR) err(rhoer1);
  av=avma;
  d1=bezout((GEN)x[2],(GEN)x[1],&x2,&y2); v1=divii((GEN)x[1],d1);
  d2=(gcmp1(d1))?gun:mppgcd(d1,(GEN)x[3]); v3=mulii(d2,v1);
  m=mulii((GEN)x[3],x2); setsigne(m,-signe(m));
  r=modii(m,v3); b3=shifti((p1=mulii(v1,r)),1);
  c3=addii(mulii((GEN)x[3],d1),mulii(r,addii((GEN)x[2],p1)));
  z=cgetg(5,t_QFR); z[1]=lmulii(v3,v1);
  z[2]=laddii((GEN)x[2],b3); z[3]=ldivii(c3,v3);
  z[4]=lshiftr((GEN)x[4],1); tetpil=avma;
  return gerepile(av,tetpil,raw? gcopy(z): redreal(z));
}

GEN
sqcompreal(GEN x)
{
  return sqcompreal0(x,0);
}

GEN
sqcomprealraw(GEN x)
{
  return sqcompreal0(x,1);
}

GEN
powrealraw(GEN x, long n, long prec)
{
  long av,tetpil;
  GEN p1,p2,y,z;

  if (n<0) err(impl,"powrealraw for negative exponents");
  if (n==1) return gcopy(x);

  z=x; av=avma; y=cgetg(5,t_QFR); y[1]=un;
  p1=subii(sqri((GEN)x[2]),shifti(mulii((GEN)x[1],(GEN)x[3]),2));
  p2=racine(p1); if (mpodd(subii(p2,(GEN)x[2]))) p2=addsi(-1,p2);
  y[2]=(long)p2; y[3]=lshifti(subii(sqri(p2),p1),-2);
  p1=cgetr(prec); y[4]=(long)p1;
  p1[1]=HIGHEXPOBIT-((prec-2)<<TWOPOTBITS_IN_LONG);
  p1[2]=0; tetpil=avma; y=gerepile(av,tetpil,gcopy(y));
  if (!n) return y;

  for (; n>1; n>>=1)
  {
    if (n&1) y=comprealraw(y,z);
    z=sqcomprealraw(z);
  }
  tetpil=avma; return gerepile(av,tetpil,comprealraw(y,z));
}

GEN
nupow(GEN x, GEN n)
{
  long av,tetpil,i,j;
  unsigned long m;
  GEN p1,p2,y,z,l;

  if (typ(n) != t_INT) err(nupower1);
  if (gcmp1(n)) return gcopy(x);
  z=x; av=avma;
  y=cgetg(4,t_QFI); y[1]=un; y[2]=mpodd((GEN)x[2]) ? un : zero;
  p1=mulii((GEN)x[1],(GEN)x[3]); p2=shifti(sqri((GEN)x[2]),-2);
  y[3]=lsubii(p1,p2);
  if (!signe(n)) { tetpil=avma; return gerepile(av,tetpil,gcopy(y)); }

  l=racine(shifti(racine((GEN)y[3]),1));
  for (i=lgefint(n)-1; i>2; i--)
  {
    for (m=n[i],j=0; j<BITS_IN_LONG; j++,m>>=1)
    {
      if (m&1) y=nucomp(y,z,l);
      z=nudupl(z,l);
    }
  }
  for (m=n[2]; m>1; m>>=1)
  {
    if (m&1) y=nucomp(y,z,l);
    z=nudupl(z,l);
  }
  tetpil=avma; y=nucomp(y,z,l);
  if (signe(n)<0 && cmpii((GEN)y[1],(GEN)y[2]) && cmpii((GEN)y[1],(GEN)y[3]))
    setsigne(y[2],-signe(y[2]));
  return gerepile(av,tetpil,y);
}

static GEN
rhoimag(GEN x)
{
  long av=avma,tetpil,s,fl,fg;
  GEN b1,c1,p1,d,z, a=(GEN)x[1], b=(GEN)x[2], c=(GEN)x[3];

  s=signe(b); setsigne(b, s? 1: 0);
  if ( (fl=cmpii(a,c)) > 0 || (fg=cmpii(a,b)) < 0 )
  {
    p1=shifti(c,1); d=dvmdii(b,p1,&b1); setsigne(b,s);
    if (s>=0)
    {
      if (cmpii(b1,c)<=0) { setsigne(d,-signe(d)); setsigne(b1,-signe(b1)); }
      else { d=subsi(-1,d); b1=subii(p1,b1); }
    }
    else
    {
      if (cmpii(b1,c)>=0) { d=addsi(1,d); b1=subii(b1,p1); }
    }
    c1=addii(a,mulii(d,shifti(subii(b,b1),-1)));
    a=c; b=b1; c=c1;
    s=signe(b); setsigne(b, s? 1: 0);
  }
  if (fl && fg && s<0) setsigne(b,s); tetpil=avma;
  z=cgetg(4,t_QFI); z[1]=lcopy(a); z[2]=lcopy(b); z[3]=lcopy(c);
  return gerepile(av,tetpil,z);
}

GEN
qfbred0(GEN x, long flag, GEN D, GEN isqrtD, GEN sqrtD)
{
  long tx=typ(x);

  if (!is_qf_t(tx)) err(rhoer1);
  if (gcmp0(D))
    D=subii(sqri((GEN)x[2]),shifti(mulii((GEN)x[1],(GEN)x[3]),2));
  switch(signe(D))
  {
    case 1: return redreal0(x,flag,D,isqrtD,sqrtD);
    case -1: if (flag && flag !=1) err(flagerr);
             return (flag&1)? rhoimag(x) : redimag(x);
  }
  return NULL; /* not reached */
}

GEN
redimag(GEN x)
{
  long av=avma,tetpil,s,fl,fg;
  GEN b1,c1,p1,d,z, a=(GEN)x[1], b=(GEN)x[2], c=(GEN)x[3];

  s=signe(b); setsigne(b, s? 1: 0);
  while ( (fl=cmpii(a,c)) > 0 || (fg=cmpii(a,b)) < 0 )
  {
    p1=shifti(c,1); d=dvmdii(b,p1,&b1); setsigne(b,s);
    if (s>=0)
    {
      if (cmpii(b1,c)<=0) { setsigne(d,-signe(d)); setsigne(b1,-signe(b1)); }
      else { d=subsi(-1,d); b1=subii(p1,b1); }
    }
    else
    {
      if (cmpii(b1,c)>=0) { d=addsi(1,d); b1=subii(b1,p1); }
    }
    c1=addii(a,mulii(d,shifti(subii(b,b1),-1)));
    a=c; b=b1; c=c1;
    s=signe(b); setsigne(b, s? 1: 0);
  }
  if (fl && fg && s<0) setsigne(b,s); tetpil=avma;
  z=cgetg(4,t_QFI); z[1]=lcopy(a); z[2]=lcopy(b); z[3]=lcopy(c);
  return gerepile(av,tetpil,z);
}

/* if sqrtD != NULL, compute Shanks' distance as well */
static GEN
rhoreal5(GEN x, GEN D, GEN isqrtD, GEN sqrtD)
{
  long s;
  GEN y,p1,nn;

  y=cgetg(5,t_QFR); y[1]=lcopy((GEN)x[3]);
  s=signe(y[1]); setsigne(y[1],1); p1=shifti((GEN)y[1],1);
  if (cmpii(isqrtD,(GEN)y[1])>=0)
    nn=divii(addii(isqrtD,(GEN)x[2]),p1);
  else
    nn=divii(addii((GEN)y[1],(GEN)x[2]),p1);
  p1=mulii(nn,p1); y[2]=lsubii(p1,(GEN)x[2]);
  setsigne(y[1],s);
  p1=shifti(subii(sqri((GEN)y[2]),D),-2);
  y[3]=ldivii(p1,(GEN)y[1]);
  if (!sqrtD)
    y[4]=x[4];
  else
    y[4] = laddrr( shiftr( mplog(absr(divrr(addir((GEN)x[2],sqrtD), subir((GEN)x[2],sqrtD)))), -1), (GEN) x[4]);
  return y;
}

static GEN
redreal5(GEN x,GEN D, GEN isqrtD, GEN sqrtD)
{
  for(;;)
  {
    if (signe(x[2]) > 0 && cmpii((GEN)x[2],isqrtD) <= 0 )
    {
      GEN p1=subii(isqrtD, shifti(absi((GEN)x[1]),1));

      if (signe(p1)<0)
      {  
	setsigne(p1,1); 
	if (cmpii((GEN)x[2],p1) >= 0) break;
      }
      else if (cmpii((GEN)x[2],p1) > 0) break;
    }
    x=rhoreal5(x,D,isqrtD,sqrtD);
  }
  return x;
}

GEN
redreal0(GEN x, long flag, GEN D, GEN isqrtD, GEN sqrtD)
{
  long av=avma,l,step, use_d;

  if (typ(x) != t_QFR) err(rhoer1);
  switch(flag)
  {
    case 0: use_d=1; step=0; break;
    case 1: use_d=1; step=1; break;
    case 2: use_d=0; step=0; break;
    case 3: use_d=0; step=1; break;
    default: err(flagerr);
  }

  if (gcmp0(D))
    D=subii(sqri((GEN)x[2]),
	    shifti(mulii((GEN)x[1],(GEN)x[3]),2));
  else if (typ(D) != t_INT) err(arither1);

  if (gcmp0(isqrtD) || (use_d && gcmp0(sqrtD)))
  {
    l=max(lg(x[4]),3);
    l=max(l,((BITS_IN_LONG-1-expo(x[4]))>>TWOPOTBITS_IN_LONG)+2);
    sqrtD=gsqrt(D,l); isqrtD=mptrunc(sqrtD);
  }
  else
  {
    if (typ(isqrtD) != t_INT) err(arither1);
    l=typ(sqrtD); if (!is_intreal_t(l)) err(arither1);
  }
  if (!use_d) sqrtD=NULL;
  if (step)
    x=rhoreal5(x,D,isqrtD,sqrtD);
  else
    x=redreal5(x,D,isqrtD,sqrtD);

  l=avma; return gerepile(av,l,gcopy(x));
}

GEN
redreal(GEN x)
{
  return redreal0(x,0,gzero,gzero,gzero);
}

GEN
rhoreal(GEN x)
{
  return redreal0(x,1,gzero,gzero,gzero);
}

GEN
redrealnod(GEN x, GEN isqrtD)
{
  return redreal0(x,2,gzero,isqrtD,gzero);
}

GEN
rhorealnod(GEN x, GEN isqrtD)
{
  return redreal0(x,3,gzero,isqrtD,gzero);
}

GEN
primeform(GEN x, GEN p, long prec)
{
  long av,tetpil,s,t,sb,sx=signe(x);
  GEN y,b,c;

  if (typ(x) != t_INT || !sx) err(arither1);
  if (gcmpgs(p,2))
  {
    y=(sx<0) ? cgetg(4,t_QFI) : cgetg(5,t_QFR);
    y[1]=lcopy(p); y[2]=lgeti(lgefint(p));
    av=avma; if (!mpsqrtmod(x,p,&b)) err(sqrter5);
    s=mod2(b); t=mod2(x);
    if (odd(s+t)) { subiiz(p,b,(GEN)y[2]); } else affii(b,(GEN)y[2]);
    c=shifti(subii(sqri((GEN)y[2]),x),-2); tetpil=avma;
    y[3]=lpile(av,tetpil,divii(c,p));
  }
  else
  {
    s=mod8(x); if (signe(x)<0) s=8-s;
    switch(s)
    {
      case 0: case 8: sb=0; break;
      case 1: sb=1; break;
      case 4: sb=2; break;
      default: err(sqrter5);
    }
    y=(sx<0) ? cgetg(4,t_QFI) : cgetg(5,t_QFR);
    y[1]=lcopy(gdeux); y[2]=lstoi(sb);
    av=avma; c=gsubsg(sb*sb,x); tetpil=avma;
    y[3]=lpile(av,tetpil,shifti(c,-3));
  }
  if (sx>0) affsr(0,(GEN)(y[4]=lgetr(prec)));
  return y;
}
