/*******************************************************************/
/*******************************************************************/
/*                                                                 */
/*               S-CLASS GROUP AND NORM SYMBOLS                    */
/*          (Denis Simon, desimon@math.u-bordeaux.fr)              */
/*                                                                 */
/*******************************************************************/
/*******************************************************************/
/* $Id: buch4.c,v 1.921 1997/03/26 00:05:38 karim Exp $ */
#include "genpari.h"

#define isvector(x) (typ(x)==t_VEC || typ(x)==t_COL)

/* fonctions techniques */
static long psquare(GEN a,GEN p);
static long lemma6(GEN pol,GEN p, long nu,GEN x);
static long lemma7(GEN pol,long nu,GEN x);
static long zpsol(GEN pol,GEN p,long nu,GEN pnu,GEN x0);

static long psquarenf(GEN nf,GEN a,GEN p);
static long psquare2nf(GEN nf,GEN a,GEN p,GEN zinit);
static long psquare2qnf(GEN nf,GEN a,GEN p, long q);
static long lemma6nf(GEN nf,GEN pol,GEN p,long nu,GEN x);
static long lemma7nf(GEN nf,GEN pol,GEN p,long nu,GEN x,GEN zinit);
static long zpsolnf(GEN nf,GEN pol,GEN p,long nu,GEN pnu,GEN x0,GEN repr,GEN zinit);

static GEN repres(GEN nf,GEN p);

static long hilb2nf(GEN nf,GEN a,GEN b,GEN p);

static void vecconcat(GEN bnf,GEN relnf,GEN concat,GEN *prod,GEN *S1,GEN *S2);

static long
psquare(GEN a,GEN p)
{
  long v;
  GEN ap;

  if (gcmp0(a) || gcmp1(a)) return 1;

  if (!cmpis(p,2))
  { 
    v=vali(a); if (v&1) return 0;
    return (smodis(shifti(a,-v),8)==1);
  }

  ap=stoi(1); v=pvaluation(a,p,&ap);
  if (v&1) return 0;
  return (kronecker(ap,p)==1);
}

static long
lemma6(GEN pol,GEN p,long nu,GEN x)
{
  long i,lambda,mu,ltop=avma;
  GEN gx,gpx;

  for (i=lgef(pol)-2,gx=(GEN) pol[i+1]; i>1; i--)
    gx=addii(mulii(gx,x),(GEN) pol[i]);
  if (psquare(gx,p)) return 1;

  for (i=lgef(pol)-2,gpx=mulis((GEN) pol[i+1],i-1); i>2; i--)
    gpx=addii(mulii(gpx,x),mulis((GEN) pol[i],i-2));

  lambda=pvaluation(gx,p,&gx);
  if (gcmp0(gpx)) mu=BIGINT; else mu=pvaluation(gpx,p,&gpx);
  avma=ltop;

  if (lambda>(mu<<1)) return 1;
  if (lambda>=(nu<<1) && mu>=nu) return 0;
  return -1;
}

static long
lemma7(GEN pol,long nu,GEN x)
{ long i,odd4,lambda,mu,mnl,ltop=avma;
  GEN gx,gpx,oddgx;

  for (i=lgef(pol)-2,gx=(GEN) pol[i+1]; i>1; i--)
    gx=addii(mulii(gx,x),(GEN) pol[i]);
  if (psquare(gx,gdeux)) return 1;

  for (i=lgef(pol)-2,gpx=gmulgs((GEN) pol[i+1],i-1); i>2; i--)
    gpx=gadd(gmul(gpx,x),gmulgs((GEN) pol[i],i-2));

  lambda=vali(gx);
  if (gcmp0(gpx)) mu=BIGINT; else mu=vali(gpx);
  oddgx=shifti(gx,-lambda);
  mnl=mu+nu-lambda;
  odd4=smodis(oddgx,4);
  avma=ltop;
  if (lambda>(mu<<1)) return 1;
  if (nu > mu)
    { if (mnl==1 && (lambda&1) == 0) return 1;
      if (mnl==2 && (lambda&1) == 0 && odd4==1) return 1;
    }
  else
    { if (lambda>=(nu<<1)) return 0;
      if (lambda==((nu-1)<<1) && odd4==1) return 0;
    }
  return -1;
}

static long
zpsol(GEN pol,GEN p,long nu,GEN pnu,GEN x0)
{
  long i,result,ltop=avma;
  GEN x,pnup;

  result = (cmpis(p,2)) ? lemma6(pol,p,nu,x0) : lemma7(pol,nu,x0);
  if (result==+1) return 1; if (result==-1) return 0;
  x=gcopy(x0); pnup=mulii(pnu,p);
  for (i=0; i<itos(p); i++)
  {
    x=addii(x,pnu);
    if (zpsol(pol,p,nu+1,pnup,x)) { avma=ltop; return 1; }
  }
  avma=ltop; return 0;
}

long
zpsoluble(GEN pol,GEN p)
/* vaut 1 si l'equation y^2=Pol(x) a une solution p-adique entiere
 * 0 sinon.
 * Les coefficients sont entiers.
 */
{
  if ((typ(pol)!=t_POL && typ(pol)!=t_INT) || typ(p)!=t_INT )
    err(typeer,"zpsoluble");
  return zpsol(pol,p,0,gun,gzero);
}

long
qpsoluble(GEN pol,GEN p)
/* vaut 1 si l'equation y^2=Pol(x) a une solution p-adique rationnelle
 * (eventuellement infinie), 0 sinon.
 * Les coefficients sont entiers.
 */
{
  if ((typ(pol)!=t_POL && typ(pol)!=t_INT) || typ(p)!=t_INT )
    err(typeer,"qpsoluble");
  if (zpsol(pol,p,0,gun,gzero)) return 1;
  return (zpsol(polrecip(pol),p,1,p,gzero));
}

static long
psquarenf(GEN nf,GEN a,GEN p)
  /* renvoie 1 si a est un carre dans ZK_p 0 sinon */
  /* seulement pour p premier avec 2 */
{
  long v,ltop=avma;
  GEN norm,ap;

  if (gcmp0(a)) return 1;
  v=idealval(nf,a,p);
  if (v&1) { avma=ltop; return 0; }
  ap=gdiv(a,gpuigs(basistoalg(nf,(GEN)p[2]),v));

  norm=gshift(idealnorm(nf,p),-1);
  ap=gmul(ap,gmodulsg(1,(GEN)p[1]));
  ap=gaddgs(gpui(ap,norm,0),-1);
  if (gcmp0(ap)) { avma=ltop; return 1; }
  ap=lift(lift(ap));
  if (idealval(nf,ap,p)>0) { avma=ltop; return 1; }
  avma=ltop; return 0;
}

static long
psquare2nf(GEN nf,GEN a,GEN p,GEN zinit)
/* renvoie 1 si a est un carre dans ZK_p 0 sinon */
/* seulement pour p divise 2 */
{
  long i,v,ltop=avma;
  GEN ap,zlog;

  if (gcmp0(a)) return 1;
  v=idealval(nf,a,p);
  if (v&1) { avma=ltop; return 0; }
  ap=gdiv(a,gmodulcp(gpuigs(gmul((GEN)nf[7],(GEN)p[2]),v),(GEN)nf[1]));

  zlog=zideallog(nf,ap,zinit);
  for (i=1; i<lg(gmael(zinit,2,2)); i++)
    if (!mpodd(gmael3(zinit,2,2,i)) && mpodd((GEN)zlog[i]))
      { avma=ltop; return 0; }
  avma=ltop; return 1;
}

static long
psquare2qnf(GEN nf,GEN a,GEN p,long q)
/* cette fonction renvoie 1 si a est un carre inversible mod P^q et 0 sinon.
 * seulement pour P divise 2, et (a,p)=1.
 */
{
  long i, ltop=avma;
  GEN zinit,zlog;

  zinit=zidealstarinit(nf,idealpows(nf,p,q));
  zlog=zideallog(nf,a,zinit);
  for (i=1; i<lg(gmael(zinit,2,2)); i++)
    if (!mpodd(gmael3(zinit,2,2,i)) && mpodd((GEN)zlog[i]))
      { avma=ltop; return 0; }
  avma=ltop; return 1;
}

static long
lemma6nf(GEN nf,GEN pol,GEN p,long nu,GEN x)
{
  long i,lambda,mu,ltop=avma;
  GEN gx,gpx;

  for (i=lgef(pol)-2,gx=(GEN) pol[i+1]; i>1; i--)
    gx=gadd(gmul(gx,x),(GEN) pol[i]);
  if (psquarenf(nf,gx,p)) { avma=ltop; return 1; }
  lambda=idealval(nf,gx,p);
  avma=ltop;

  for (i=lgef(pol)-2,gpx=gmulgs((GEN) pol[i+1],i-1); i>2; i--)
    gpx=gadd(gmul(gpx,x),gmulgs((GEN) pol[i],i-2));
  if (!gcmp0(gpx)) mu=idealval(nf,gpx,p); else mu=BIGINT;
  avma=ltop;

  if (lambda>(mu<<1)) return 1;
  if ((lambda >= (nu<<1))  && (mu >= nu)) return 0;
  return -1;
}

static long
lemma7nf(GEN nf,GEN pol,GEN p,long nu,GEN x,GEN zinit)
{
  long res,i,lambda,mu,q,ltop=avma;
  GEN gx,gpx;

  for (i=lgef(pol)-2, gx=(GEN) pol[i+1]; i>1; i--)
    gx=gadd(gmul(gx,x),(GEN) pol[i]);
  if (psquare2nf(nf,gx,p,zinit)) { avma=ltop; return 1; }
  lambda=idealval(nf,gx,p);

  for (i=lgef(pol)-2,gpx=gmulgs((GEN) pol[i+1],i-1); i>2; i--)
    gpx=gadd(gmul(gpx,x),gmulgs((GEN) pol[i],i-2));
  if (!gcmp0(gpx)) mu=idealval(nf,gpx,p); else mu=BIGINT;

  if (lambda>(mu<<1)) { avma=ltop; return 1; }
  if (nu > mu)
  {
    if (lambda&1) { avma=ltop; return -1; }
    q=mu+nu-lambda; res=1;
  }
  else
  {
    if (lambda>=(nu<<1)) { avma=ltop; return 0; }
    if (lambda&1) { avma=ltop; return -1; }
    q=(nu<<1)-lambda; res=0;
  }
  if (q > itos((GEN) p[3])<<1) { avma=ltop; return -1; }
  if ( psquare2qnf(nf,gdiv(gx,
                          gmodulcp(gpuigs(gmul((GEN)nf[7],(GEN)p[2]),
			                  lambda),(GEN)nf[1])),
		      p,q) )
    { avma=ltop; return res; }
  avma=ltop; return -1;
}

static long
zpsolnf(GEN nf,GEN pol,GEN p,long nu,GEN pnu,GEN x0,GEN repr,GEN zinit)
{
  long i,result,ltop=avma;
  GEN pnup;

  nf=checknf(nf);
  if (cmpis((GEN) p[1],2))
    result=lemma6nf(nf,pol,p,nu,x0);
  else
    result=lemma7nf(nf,pol,p,nu,x0,zinit);
  if (result==+1) return 1; if (result==-1) return 0;
  pnup=gmul(pnu,gmodulcp(gmul((GEN) nf[7],(GEN) p[2]),(GEN) nf[1]));
  nu++;
  for (i=1; i<lg(repr); i++)
    if (zpsolnf(nf,pol,p,nu,pnup,gadd(x0,gmul(pnu,(GEN)repr[i])),repr,zinit))
    { avma=ltop; return 1; }
  avma=ltop; return 0;
}

static GEN
repres(GEN nf,GEN p)
/* calcule un systeme de representants Zk/p */
/* nf est donne par initalg */
{
  long i,j,k,f,pp,ppf,ppi,lbot,ltop=avma;
  GEN mat,fond,rep;

  fond=cgetg(1,t_VEC);
  mat=idealhermite(nf,p);
  for (i=1; i<lg(mat); i++)
    if (!gcmp1(gmael(mat,i,i)))
      fond=concat(fond,gmael(nf,7,i));
  f=lg(fond)-1;
  pp=itos((GEN) p[1]);
  for (i=1,ppf=1; i<=f; i++) ppf*=pp;
  rep=cgetg(ppf+1,t_VEC);
  rep[1]=zero; ppi=1;
  for (i=0; i<f; i++,ppi*=pp)
    for (j=1; j<pp; j++)
      for (k=1; k<=ppi; k++)
	rep[j*ppi+k]=ladd((GEN) rep[k],gmulsg(j,(GEN) fond[i+1]));
  lbot=avma;
  return gerepile(ltop,lbot,gmodulcp(rep,(GEN) nf[1]));
}

long
qpsolublenf(GEN nf,GEN pol,GEN p)
  /* =1 si l'equation y^2 = z^deg(pol) * pol(x/z) a une solution rationnelle
   *    p-adique (eventuellement (1,y,0) = oo)
   * =0 sinon.
   * Les coefficients de pol doivent etre des entiers de nf.
   * p est un ideal premier sous forme primedec.
   */
{
  GEN repr,zinit;
  long ltop=avma;

  if (gcmp0(pol)) return 1;
  if (typ(pol)!=t_POL)
    err (talker,"not a polynom in qpsolublenf");
  if (typ(p)!=t_VEC  || lg(p)!=6)
    err (talker,"not a prime ideal in qpsolublenf");
  nf=checknf(nf);

  if (cmpis((GEN) p[1],2))
  {
    if (psquarenf(nf,(GEN) pol[2],p)) return 1;
    if (psquarenf(nf, leading_term(pol),p)) return 1;
    zinit=gzero;
  }
  else
  {
    zinit=zidealstarinit(nf,idealpows(nf,p,1+2*idealval(nf,gdeux,p)));
    if (psquare2nf(nf,(GEN) pol[2],p,zinit)) return 1;
    if (psquare2nf(nf, leading_term(pol),p,zinit)) return 1;
  }
  repr=repres(nf,p);
  if (zpsolnf(nf,pol,p,0,gun,gzero,repr,zinit)) { avma=ltop; return 1; }
  if (zpsolnf(nf,polrecip(pol),p,1,gmodulcp(gmul((GEN) nf[7],(GEN) p[2]),(GEN) nf[1]),gzero,repr,zinit))
    { avma=ltop; return 1; }

  avma=ltop; return 0;
}

long
zpsolublenf(GEN nf,GEN pol,GEN p)
  /* =1 si l'equation y^2 = pol(x) a une solution entiere p-adique
   * =0 sinon.
   * Les coefficients de pol doivent etre des entiers de nf.
   * p est un ideal premier sous forme primedec.
  */
{
  GEN repr,zinit;
  long ltop=avma;

  if (gcmp0(pol)) return 1;
  if (typ(pol)!=t_POL)
    err (talker,"not a polynom in zpsolublenf");
  if (typ(p)!=t_VEC || lg(p)!=6)
    err (talker,"not a prime ideal in zpsolublenf");
  nf=checknf(nf);

  if (cmpis((GEN)p[1],2))
  {
    if (psquarenf(nf,(GEN) pol[2],p)) return 1;
    zinit=gzero;
  }
  else
  {
    zinit=zidealstarinit(nf,idealpows(nf,p,1+2*idealval(nf,gdeux,p)));
    if (psquare2nf(nf,(GEN) pol[2],p,zinit)) return 1;
  }
  repr=repres(nf,p);
  if (zpsolnf(nf,pol,p,0,gun,gzero,repr,zinit))
    { avma=ltop; return 1; }

  avma=ltop; return 0;
}

static long
hilb2nf(GEN nf,GEN a,GEN b,GEN p)
{
  GEN pol;
  long ltop=avma;

  a=lift(a);b=lift(b);
  pol=polx[0]; pol=gadd(gmul(a,gsqr(pol)),b);
  if (qpsolublenf(nf,pol,p)) { avma=ltop; return 1; }
  avma=ltop; return -1;
}

static GEN
nfmodid2(GEN nf,GEN x,GEN ideal)
/* ideal doit etre sous la forme primedec */
{
  x=lift(x);
  x=gmod(x,lift(basistoalg(nf,(GEN)ideal[2])));
  return (gmul(x,gmodulcp(gun,(GEN)ideal[1])));
}

long
nfhilbertp(GEN nf,GEN a,GEN b,GEN p)
/* calcule le symbole de Hilbert quadratique local (a,b)_p
 * en l'ideal premier p du corps nf,
 * a et b sont des elements non nuls de nf, sous la forme
 * de polymods ou de polynomes, et p renvoye par primedec.
 */
{
  GEN aux,aux2;
  long ta=typ(a),tb=typ(b),alpha,beta,sign,rep,ltop=avma;

  if ((ta!=t_INT && ta!=t_POLMOD && ta!=t_POL) 
   || (tb!=t_INT && tb!=t_POLMOD && tb!=t_POL))
    err (typeer,"nfhilbertp");
  if (gcmp0(a) || gcmp0(b))
    err (talker,"0 argument in nfhilbertp");
  if (typ(p)!=t_VEC || lg(p)!=6)
    err (talker,"not a prime ideal in nfhilbertp");
  nf=checknf(nf);

  if (!cmpis((GEN) p[1],2)) return hilb2nf(nf,a,b,p);

  nf=checknf(nf);
  if (ta != t_POLMOD) a=gmodulcp(a,(GEN)nf[1]);
  if (tb != t_POLMOD) b=gmodulcp(b,(GEN)nf[1]);
  
  alpha=idealval(nf,a,p); beta=idealval(nf,b,p);
  if ((alpha&1) == 0 && (beta&1) == 0) { avma=ltop; return 1; }
  aux2=shifti(idealnorm(nf,p),-1);
  if (alpha&1 && beta&1 && mpodd(aux2)) sign=1; else sign=-1;
  aux=nfmodid2(nf,gdiv(gpuigs(a,beta),gpuigs(b,alpha)),p); /* ?????? */
  aux=gaddgs(gpui(aux,aux2,0),sign);
  aux=lift(lift(aux));
  if (gcmp0(aux)) rep=1; else rep=(idealval(nf,aux,p)>=1);
  avma=ltop;
  if (rep) return 1; else return -1;
}

long
nfhilbert(GEN nf,GEN a,GEN b)
/* calcule le symbole de Hilbert quadratique global (a,b):
 * =1 si l'equation X^2-aY^2-bZ^2=0 a une solution non triviale,
 * =-1 sinon,
 * a et b doivent etre non nuls.
 */
{
  long ta=typ(a),tb=typ(b),r1,i,ltop=avma;
  GEN S,al,bl;

  nf=checknf(nf);
  if ((ta!=t_INT && ta!=t_POLMOD && ta!=t_POL)
   || (tb!=t_INT && tb!=t_POLMOD && tb!=t_POL))
    err (typeer,"nfhilbert");
  if (gcmp0(a) || gcmp0(b))
    err (talker,"0 argument in nfhilbert");

  al=lift(a); bl=lift(b);
 /* solutions locales aux places infinies reelles */
  r1=itos(gmael(nf,2,1));
  for (i=1; i<=r1; i++)
    if (signe(poleval(al,gmael(nf,6,i)))==-1)
      if (signe(poleval(bl,gmael(nf,6,i)))==-1)
        { 
	  avma=ltop;
	  if (DEBUGLEVEL>=4)
	  {
	    fprintferr("nfhilbert not soluble at a real place\n");
	    flusherr();
	  }
	return -1; }

  if (ta!=t_POLMOD) a=gmodulcp(a,(GEN)nf[1]);
  if (tb!=t_POLMOD) b=gmodulcp(b,(GEN)nf[1]);

      /* solutions locales aux places finies (celles qui divisent 2ab)*/

  S=(GEN) idealfactor(nf,gmul(gmulsg(2,a),b))[1];
  for (i=lg(S)-1; i>1; i--) /* d'apres la formule du produit on peut eviter un premier */
    if (nfhilbertp(nf,a,b,(GEN) S[i])==-1)
    { 
      avma=ltop;
      if (DEBUGLEVEL >=4)
      {
	fprintferr("nfhilbert not soluble at finite place: ");
	outerr((GEN)S[i]); flusherr();
      }
      return -1;
    }

  avma=ltop; return 1;
}

long
nfhilbert0(GEN nf,GEN a,GEN b,GEN p)
{
  if (gcmp0(p)) return nfhilbert(nf,a,b);
  return nfhilbertp(nf,a,b,p);
}

GEN
bnfsunit(GEN bnf,GEN S,long PREC)
  /* donne un vecteur "res" avec 7 composantes:
   * res[1] est un systeme de generateurs des s-unites (entieres) qui ne sont
   * pas des unites,exprimees dans la Z-base;
   * res[2] contient les coordonnees de res[1] sur les ideaux de S
   * (sous forme HNF);
   * res[3] donne les plongements logarithmiques de res[1];
   * res[4] est le s-regulateur, c'est a dire le produit du regulateur,
   * du determinant de res[2], et des log des normes des ideaux de S;
   * res[5] donne la structure du s-groupe de classe;
   * res[6] est la liste s;
   * bnf est le resultat de buchinit.
   * S est une liste d'ideaux premiers sous la forme de primedec.
   */
{
  long d,i,j,l,ls,lpow,ltop=avma,lbot;
  GEN bnf7,classgp;
  GEN M,Mp,aux,v,xd;
  GEN U,U1,H,mask;
  GEN sunites,s_groupe,sunites_entieres;
  GEN card,news_groupe,sreg,res,pow;

  checkbnf(bnf);
  if (!isvector(S)) err(typeer,"bnfsunit");
  classgp=gmael(bnf,8,1); bnf7=(GEN) bnf[7];
  l=lg(classgp[2]); d=lg(bnf7[1]);

  ls=lg(S);
  res=cgetg(7,t_VEC);
  res[1]=(long) S;
  res[2]=lgetg(1,t_VEC);
  res[3]=lgetg(1,t_VEC);
  res[4]=un;
  res[5]=(long) classgp;
  res[6]=(long) S;

    /* matrice des relations du groupe de classes de bnf */		
  M=diagonal((GEN) classgp[2]);
  Mp=cgetg(l,t_VEC);
  for (i=1; i<l; i++)
  {
    aux=isprincipalgen(bnf,idealpow(bnf,gmael(classgp,3,i),gmael(classgp,2,i)));
    Mp[i]=linvmod(gmul((GEN) bnf7[7],(GEN) aux[2]),(GEN) bnf7[1]);
  }

 /* matrice des nouvelles relations dans le groupe de S-classes
  * = ideaux de S exprimes dans le groupe de classes
  */
  for (i=1; i<ls; i++)
  {
    aux=isprincipalgen(bnf,(GEN) S[ls-i]);
    M=concat((GEN) aux[1],M);
    Mp=concat(gmul((GEN) bnf7[7],(GEN) aux[2]),Mp);
  }

    /* calcul des s-unites */
  H=hnfall(M);
  U=(GEN) H[2];

  sunites=cgetg(ls+l-1,t_VEC);
  for (i=1; i<ls+l-1; i++)
  {
    aux=gun; 
    for (j=1; j<ls+l-1; j++)
      if (!gcmp0(gmael(U,i,j)))
	aux=gmul(aux,gpui((GEN) Mp[j],gmael(U,i,j),0));
    sunites[i]=lmod(aux,(GEN) bnf7[1]);
  }

      /* calcul du s-groupe de classes */
  card=gun;
  if (lg(H)>2) if (lg(H[1])>1)
  {
    s_groupe=smith2((GEN) H[1]);
    news_groupe=cgetg(4,t_VEC);
    news_groupe[1]=ldet((GEN)s_groupe[3]);
    if (gcmp1((GEN)news_groupe[1]))
      { news_groupe[2]=lgetg(1,t_VEC);
      news_groupe[3]=lgetg(1,t_VEC); }
    else
    {
      news_groupe[2]=(long) smith((GEN)s_groupe[3]);
      news_groupe[3]=lgetg(lg(s_groupe[1]),t_VEC);
      for(i=1,lpow=1;i<lg(news_groupe[2]);i++)
	if (!gcmp1(gmael(news_groupe,2,i))) lpow++;
        else setlg(news_groupe[2],lpow);
      news_groupe[3]=lgetg(lpow,t_VEC);
      pow=ginv((GEN)s_groupe[1]);
      for(i=1;i<lpow;i++)
      {
	for(j=1,aux=gun;j<lg(s_groupe[1]);j++)
	  aux=idealmul(bnf,aux,idealpow(bnf,gmael4(bnf,8,1,3,j),
					    gmael(pow,i,j)));
	mael(news_groupe,3,i)=(long)aux;
      }
    }
    res[5]=(long) news_groupe;
    card=dethnf((GEN)H[1]);
  }

      /* expression des s-unites comme entiers de bnf */
  sunites_entieres=cgetg(ls,t_VEC);
  if (ls>1)
  {
    mask=subis(shifti(gun,ls-1),1);
    U=matextract(U,mask,mask);
    U1=(GEN) hnfall(U)[2]; res[2]=lmul(U,U1);
    /* la matrice U1 contient les coord. des s-unites sur les ideaux de S */
    v=cgetg(d-2,t_COL); xd=gpuigs((GEN) polx[0], d);
    for (i=1; i<ls; i++)
    { 
      aux=gun;
      for (j=1; j<ls; j++)
        aux=gmul(aux,gpui((GEN) sunites[j],gmael(U1,i,j),0));
      aux=gadd(gmod(aux,(GEN) bnf7[1]),xd);
      for (j=1; j<d-2; j++) v[j]=aux[j+1];
      sunites_entieres[i]=lmul((GEN) bnf7[8],v);
    }
    res[1]=(long) sunites_entieres;
    res[3]=llog(conjvec(basistoalg(bnf,sunites_entieres),PREC),PREC);
  }

    /* calcul du s-regulateur */
  sreg=gmul(gmael(bnf,8,2),card);
  for (i=1; i<ls; i++) sreg=gmul(sreg,glog(gmael(S,i,1),PREC));
  res[4]=(long) sreg;

  lbot=avma; return gerepile(ltop,lbot,gcopy(res));
}

GEN
bnfissunit(GEN bignf,GEN suni,GEN x)
/* cette fonction est l'equivalent de isunit, sauf qu'elle donne le resultat
 * avec des s-unites: si x n'est pas une s-unite alors issunit=[]~;
 * si x est une s-unite alors
 * x=\prod_{i=0}^r {e_i^issunit[i]}*prod{i=r+1}^{r+s} {s_i^issunit[i]}
 * ou les e_i sont les unites du corps (comme dans isunit)
 * et les s_i sont les s-unites calculees par sunit (dans le meme ordre).
 */
{
  long i,ltop=avma,lbot;
  GEN S,v,x1,mat,aux,sunitalg;
  GEN res=cgetg(1,t_COL);

  if (gcmp0(x)) return res;
  checkbnf(bignf);
  if (!isvector(suni) || lg(suni)!=7) err (typeer,"bnfissunit");
  
  S=(GEN) suni[6];
  if (lg(S)==1)
    { lbot=avma; return gerepile(ltop,lbot,isunit(bignf,x)); }
  v=cgetg(lg(S),t_COL);
  for (i=1; i<lg(S); i++)
    v[i]=lstoi(idealval(bignf,x,(GEN) S[i]));
  mat=gmul(ginv((GEN) suni[2]),v);
  if (gcmp1(denom(mat)))
  {
    sunitalg=basistoalg(bignf,(GEN) suni[1]);
    x1=gcopy(x);
    for (i=1; i<lg(S); i++)
      x1=gmul(x1,gpuigs((GEN) sunitalg[i],-itos((GEN)mat[i])));
    aux=isunit(bignf,x1);
    if (lg(aux)>1)
      res=concat(aux,mat);
  }
  lbot=avma; return gerepile(ltop,lbot,gcopy(res));
}

static void
vecconcat(GEN bnf,GEN relnf,GEN vec,GEN *prod,GEN *S1,GEN *S2)
{
  long i;

  for (i=1; i<lg(vec); i++)
    if (signe(resii(*prod,(GEN)vec[i])))
    {
       *prod=mulii(*prod,(GEN)vec[i]);
       *S1=concat(*S1,primedec(bnf,(GEN)vec[i]));
       *S2=concat(*S2,primedec(relnf,(GEN)vec[i]));
    }
}

GEN
rnfisnorm(GEN bnf,GEN ext,GEN x,long flag,long PREC)

/* bnf est le corps de base (buchinitfu).
 * ext definit l'extension relative:
 * ext[1] est une equation relative du corps,
 * telle qu'une de ses racines engendre le corps sur Q.
 * ext[2] exprime le generateur (y) du corps de base,
 * en fonction de la racine (x) de ext[1],
 * ext[3] est le buchinitfu (sur Q) de l'extension.

 * si flag=0 c'est qu'on sait a l'avance que l'extension est galoisienne,
 * et dans ce cas la reponse est exacte.
 * si flag>0 alors on ajoue dans S tous les ideaux qui divisent p<=flag.
 * si flag<0 alors on ajoute dans S tous les ideaux qui divisent -flag.

 * la reponse est un vecteur v a 2 composantes telles que
 * x=N(v[1])*v[2].
 * x est une norme ssi v[2]=1.
 */

{
  long lgsunitrelnf,i,lbot, ltop = avma;
  GEN resultat = cgetg(3,t_VEC), S1 = cgetg(1,t_VEC), S2 = cgetg(1,t_VEC);
  GEN prod = gun;
  GEN relnf,aux,vec,tors,xnf,H,Y;
  GEN M,A,suni,sunitrelnf,sunitnormnf;

  checkbnf(bnf);
  if (!isvector(ext) || lg(ext)!=4)
    err (typeer,"bnfisnorm");
  
  resultat[1]=(long) x; resultat[2]=un;
  if (gcmp0(x) || gcmp1(x))
    { lbot=avma; return gerepile(ltop,lbot,gcopy(resultat)); }
  if (gcmp_1(x) && (degree((GEN)ext[1])&1))
    { lbot=avma; return gerepile(ltop,lbot,gcopy(resultat)); }

  relnf = (GEN)ext[3];

/* construction de l'ensemble S des ideaux
   qui interviennent dans les solutions */

  if (!gcmp1(gmael3(relnf,8,1,1)))
  {
    GEN genclass=gmael3(relnf,8,1,3);
    vec=cgetg(1,t_VEC);
    for(i=1;i<lg(genclass);i++)
      if (!gcmp1(ggcd(gmael4(relnf,8,1,2,i), stoi(degree((GEN)ext[1])))))
        vec=concat(vec,(GEN)factor(gmael3(genclass,i,1,1))[1]);
    vecconcat(bnf,relnf,vec,&prod,&S1,&S2);
  }

  if (flag>1)
  {
    for (i=2; i<=flag; i++)
      if (isprime(stoi(i)) && signe(resis(prod,i)))
      {
	prod=mulis(prod,i);
	S1=concat(S1,primedec(bnf,stoi(i)));
	S2=concat(S2,primedec(relnf,stoi(i)));
      }
  }
  else if (flag<0)
    vecconcat(bnf,relnf,(GEN)factor(stoi(-flag))[1],&prod,&S1,&S2);

  if (flag)
  {
    GEN normdiscrel=divii(gmael(relnf,7,3),
			  gpuigs(gmael(bnf,7,3),lg(ext[1])-3));
    vecconcat(bnf,relnf,(GEN) factor(absi(normdiscrel))[1],
	      &prod,&S1,&S2);
  }
  vec=(GEN) idealfactor(bnf,x)[1];
  for (i=1; i<lg(vec); i++)
    if (signe(resii(prod,gmael(vec,i,1))))
    {
      aux=cgetg(2,t_VEC);
      aux[1]=vec[i];
      S1=concat(S1,aux);
    }
  xnf=lift(x);
  xnf=gsubst(xnf,varn(xnf),(GEN)ext[2]);
  vec=(GEN) idealfactor(relnf,xnf)[1];
  for (i=1; i<lg(vec); i++)
    if (signe(resii(prod,gmael(vec,i,1))))
    {
      aux=cgetg(2,t_VEC);
      aux[1]=vec[i];
      S2=concat(S2,aux);
    }

  resultat[1]=un; resultat[2]=(long)x;
  tors=cgetg(2,t_VEC); tors[1]=mael3(relnf,8,4,2);

/* calcul sur les S-unites */

  suni=bnfsunit(bnf,S1,PREC);
  A=lift(bnfissunit(bnf,suni,x));
  sunitrelnf=(GEN) bnfsunit(relnf,S2,PREC)[1];
  if (lg(sunitrelnf)>1)
    {
      sunitrelnf=lift(basistoalg(relnf,sunitrelnf));
      sunitrelnf=concat(tors,sunitrelnf);
    }
  else sunitrelnf=tors;
  aux=(GEN)relnf[8];
  if (lg(aux)>=6) aux=(GEN)aux[5];
  else
  {
    aux=(GEN)buchfu(relnf);
    if(gcmp0((GEN)aux[2]))
      err(talker,"please increase precision to have units in bnfisnorm");
    else aux=(GEN)aux[1];
  }
  if (lg(aux)>1)
    sunitrelnf=concat(gtrans(aux),sunitrelnf);
  lgsunitrelnf=lg(sunitrelnf);
  M=cgetg(lgsunitrelnf+1,t_MAT);
  sunitnormnf=cgetg(lgsunitrelnf,t_VEC);
  for (i=1; i<lgsunitrelnf; i++)
    {
      sunitnormnf[i]=lnorm(gmodulcp((GEN) sunitrelnf[i],(GEN)ext[1]));
      M[i]=(long) lift(bnfissunit(bnf,suni,(GEN) sunitnormnf[i]));
    }
  M[lgsunitrelnf]=lgetg(lg(A),t_COL);
  for (i=1; i<lg(A); i++) mael(M,lgsunitrelnf,i)=zero;
  mael(M,lgsunitrelnf,lg(mael(bnf,7,6))-1)=mael3(bnf,8,4,1);
  H=hnfall(M); Y=inverseimage(gmul(M,(GEN) H[2]),A);
  Y=gmul((GEN) H[2],Y);
  for (aux=(GEN)resultat[1],i=1; i<lgsunitrelnf; i++)
    aux=gmul(aux,gpuigs(gmodulcp((GEN) sunitrelnf[i],(GEN)ext[1]),
                        itos(gfloor((GEN)Y[i]))));
  resultat[1]=(long)aux;
  resultat[2]=ldiv(x,gnorm(gmodulcp(lift(aux),(GEN)ext[1])));

  lbot=avma; return gerepile(ltop,lbot,gcopy(resultat));
}

GEN
bnfisnorm(GEN bnf,GEN x,long flag,long PREC)
{
  long ltop = avma, lbot;
  GEN ext = cgetg(4,t_VEC);

  checkbnf(bnf);
  ext[1] = mael(bnf,7,1);
  ext[2] = zero;
  ext[3] = (long) bnf;
  bnf = buchinitfu(polx[MAXVARN],NULL,NULL,0); lbot = avma;
  return gerepile(ltop,lbot,rnfisnorm(bnf,ext,x,flag,PREC));
}

/*
  else
  {
    flagell=0;
    phiell=cyclo(ell,0);
    facphiell=nffactor(nf,phiell);
    polrel1=gmael(facphiell,1,1);
    degK1=degree(polrel1);
    d=(ell-1)/degK1;
    p2=rnfequationall(nf,polrel1,1);
    polabs1=(GEN)p2[1];
    uth=(GEN)p2[2];
    gk=(GEN)p2[3];
    p2=polredabs2(polabs1,prec);
    polabs2=(GEN)p2[1];
    u1=(GEN)p2[2];
    uth2=gsubst(lift(uth),0,u1);
    a2inv=polymodrecip(u1);
    g=gener(gell);
    gd=lift(gpuigs(g,d));
    ku=gmul(gk,uth2);
    tauu2=gsubst(lift(a2inv),0,gadd(gpui(gsub(u1,ku),gd,0),ku));
    polabs=gsubst(polabs2,0,polx[vnf]);
    tauu2=gmodulcp(gsubst(lift(tauu2),0,polx[vnf]),polabs);
    bnf1=bnfinit0(polabs,gzero,1,prec);
    degabs=degnf*degK1;
    rnf=rnfinitalg(nf,polrel1,prec);
    ideallift=rnfidealup(rnf,ideal);
    bnr1=buchrayinitgen(bnf1,ideallift,prec);
    classgroup1=(GEN)bnr1[5];
    cyclic1=(GEN)classgroup1[2];
    D1=diagonal(cyclic1);
    generator1=(GEN)classgroup1[3];
    bid1=(GEN)bnr1[2];    
    lx1=lg(cyclic1);
    m=cgetg(lx1,t_MAT);
    for (j=1; j<lx1; j++)
    {
      m[j]=(long)isprincipalray(bnr,rnfidealnorm(rnf,(GEN)generator1[j]));
    }
    p1=concat(m,subgroup);
    bigu=(GEN)hnfall(p1)[2];
    mu1=cgetg(lx1,t_MAT);
    for (j=1; j<lx1; j++)
    {
      p2=cgetg(lx1,t_COL); mu1[j]=(long)p2;
      for (i=1; i<lx1; i++) p2[i]=coeff(bigu,i,j);
    }
    subgroup1=hnf(concat(mu1,D1));
    p1=conductor(bnr1,subgroup1,2,prec);
    bnr1=(GEN)p1[2]; subgroup1=(GEN)p1[3];
    classgroup1=(GEN)bnr1[5];
    cyclic1=(GEN)classgroup1[2];
    D1=diagonal(cyclic1);
    generator1=(GEN)classgroup1[3];
    bid1=(GEN)bnr1[2];    
    lx1=lg(cyclic1);
    module1=(GEN)bid1[1];
    ideal1=(GEN)module1[1];
  }
*/

/* si all!=0, donne toutes les equations correspondant a un sousgroupe
   de determinant all (i.e. de degre all) */
GEN
rnfkummer(GEN bnr, GEN subgroup, long all, long prec)
{
  long r1,vnf,degnf,ell,j,i,av=avma,tetpil;
  long nbgenclK,lprm,lprl,lSm,lSml1,lSml2,lSl,lSp,lastindex,nbvunit;
  long kmax,nbgenbid,lastbid,nbcol,nblig,k,lker,flag,cmpt,llistx,e,vp,v1;

  GEN bnf,nf,polnf,p1,classgroup,cyclic,generator,bid,module,ideal,arch;
  GEN D,gell,wk,p2;
  GEN clK,cyclicK,genK,listgamma,listalpha,fa,prm,expom,Sm,Sml1,Sml2,Sl;
  GEN pp,p,prl,Sp,vecbeta,matexpo,unmodell,vunit,id,pr,vsml2,p3,p4;
  GEN cycpro,cl,cycbid,munit,vecmunit2,msign,archartif,vecp1,kermat,listx;
  GEN yinit,y,listal,group;

  checkbnrgen(bnr); bnf=(GEN)bnr[1];
  nf=(GEN)bnf[7]; r1=itos(gmael(nf,2,1));
  polnf=(GEN)nf[1]; vnf=varn(polnf);
  if (vnf==0) err(talker,"main variable in kummer must not be x");
  p1=conductor(bnr,all ? gzero : subgroup,2,prec);
  bnr=(GEN)p1[2]; if(!all) subgroup=(GEN)p1[3];
  classgroup=(GEN)bnr[5];
  cyclic=(GEN)classgroup[2];
  generator=(GEN)classgroup[3];
  bid=(GEN)bnr[2]; module=(GEN)bid[1];
  ideal=(GEN)module[1]; arch=(GEN)module[2];
  D=diagonal(cyclic);
  if ((!all)&&gcmp0(subgroup)) subgroup=D;
  degnf=lgef(polnf)-3;
  gell=all ? stoi(all) : det(subgroup);
  if (!isprime(gell))
    err(impl,"kummer for non prime relative degree");
  ell=itos(gell);
  wk=gmael3(bnf,8,4,1);
  if (!gdivise(wk,gell)) 
  {
    err(impl,"difficult kummer");
  }
  clK=gmael(bnf,8,1); cyclicK=(GEN)clK[2]; genK=(GEN)clK[3];
  nbgenclK=lg(genK)-1;
  listgamma=cgetg(nbgenclK+1,t_VEC); listalpha=cgetg(nbgenclK+1,t_VEC);
  p1=gmul(gell,ideal);
  for (i=1; i<=nbgenclK; i++)
  {
    p2=basistoalg(nf,idealcoprime(nf,(GEN)genK[i],p1));
    listgamma[i]=(long)ginv(p2);
    p2=idealpow(nf,idealmul(nf,(GEN)genK[i],p2),(GEN)cyclicK[i]);
    listalpha[i]=(long)basistoalg(nf,(GEN)isprincipalgen(bnf,p2)[2]);
  }
/* a ce stade, module est egal au conducteur. */
  fa=idealfactor(nf,ideal);
  prm=(GEN)fa[1]; expom=(GEN)fa[2]; lprm=lg(prm)-1;
  Sm=cgetg(1,t_VEC);Sml1=cgetg(1,t_VEC);Sml2=cgetg(1,t_VEC);Sl=cgetg(1,t_VEC);
  for (i=1; i<=lprm; i++)
  {
    pp=cgetg(2,t_VEC); pp[1]=prm[i];
    p=(GEN)((GEN)pp[1])[1]; e=itos((GEN)((GEN)pp[1])[3]);
    if (!gegal(p,gell))
    {
      Sm=concat(Sm,pp);
      if (!gcmp1((GEN)expom[i])) err(talker,"bug1 in kummer");
    }
    else
    {
      vp=itos((GEN)expom[i]);
      if ((vp-1)*(ell-1)==ell*e) Sml1=concat(Sml1,pp);
      else Sml2=concat(Sml2,pp);
    }
  }
  prl=primedec(nf,gell); lprl=lg(prl)-1;
  for (i=1; i<=lprl; i++)
  {
    if (!idealval(nf,ideal,(GEN)prl[i]))
    {
      pp=cgetg(2,t_VEC); pp[1]=prl[i];
      Sl=concat(Sl,pp);
    }
  }
  lSm=lg(Sm)-1; lSml1=lg(Sml1)-1; lSml2=lg(Sml2)-1; lSl=lg(Sl)-1;
  Sp=concat(Sm,Sml1); lSp=lSm+lSml1;
  vecbeta=cgetg(lSp+1,t_VEC); matexpo=cgetg(lSp+1,t_MAT);
  for (j=1; j<=lSp; j++)
  {
    p1=isprincipalgen(bnf,(GEN)Sp[j]);
    p2=basistoalg(nf,(GEN)p1[2]);
    for (i=1; i<=nbgenclK; i++) p2=gmul(p2,gpui((GEN)listgamma[i],gmael(p1,1,i),0));
    matexpo[j]=(long)p1[1];
    vecbeta[j]=(long)p2;
  }
  unmodell=gmodulcp(gun,gell);
  matexpo=gmul(unmodell,matexpo);
  i=1; while ((i<=nbgenclK)&&(gdivise((GEN)cyclicK[i],gell))) i++;
  lastindex=i-1;
  vunit=gmodulcp(concat(gmael(bnf,8,5),gmael3(bnf,8,4,2)),(GEN)nf[1]);
  for (i=1; i<=lastindex; i++) vunit=concat(vunit,(GEN)listalpha[i]);
  nbvunit=lg(vunit)-1;
  id=idmat(degnf);
  for (i=1; i<=lSl; i++)
  {
    pr=(GEN)Sl[i];
    id=idealmul(nf,idealpow(nf,pr,stoi((ell*itos((GEN)pr[3]))/(ell-1))),id);
  }
  vsml2=cgetg(lSml2+1,t_VEC);
  for (i=1; i<=lSml2; i++)
  {
    pr=(GEN)Sml2[i];
    v1=idealval(nf,ideal,pr);
    if (v1<2) err(talker,"bug4 in kummer");
    kmax=(itos((GEN)pr[3])*ell)/(ell-1)+1-v1;
    p1=idealpow(nf,pr,stoi(kmax));
    id=idealmul(nf,id,p1);
    p2=idealmul(nf,p1,pr);
    p3=zidealstarinitgen(nf,p2);
    vsml2[i]=(long)p3;
    cycpro=gmael(p3,2,2);
    j=1; while ((j<lg(cycpro))&&(gdivise((GEN)cycpro[j],gell))) j++;
    if (j<lg(cycpro)) err(talker,"bug5 in kummer");
  }
  bid=zidealstarinitgen(nf,id);
  cl=(GEN)bid[2]; cycbid=(GEN)cl[2]; nbgenbid=lg(cycbid)-1;
  i=1; while ((i<=nbgenbid)&&(gdivise((GEN)cycbid[i],gell))) i++;
  lastbid=i-1;
  nbcol=nbvunit+lSp; nblig=lastbid+lastindex;
  munit=cgetg(nbcol+1,t_MAT); vecmunit2=cgetg(lSml2+1,t_VEC);
  msign=cgetg(nbcol+1,t_MAT);
  for (k=1; k<=lSml2; k++) vecmunit2[k]=lgetg(nbcol+1,t_MAT);
  archartif=cgetg(r1+1,t_VEC); for (j=1; j<=r1; j++) archartif[j]=un;
  for (j=1; j<=nbvunit; j++)
  {
    p1=zideallog(nf,(GEN)vunit[j],bid);
    p2=cgetg(nblig+1,t_COL); munit[j]=(long)p2;
    for (i=1; i<=lastbid; i++) p2[i]=p1[i];
    for (   ; i<=nblig; i++) p2[i]=zero;
    for (k=1; k<=lSml2; k++)
      ((GEN)vecmunit2[k])[j]=(long)zideallog(nf,(GEN)vunit[j],(GEN)vsml2[k]);
    msign[j]=(long)zsigne(nf,(GEN)vunit[j],archartif,r1);
  }
  p1=cgetg(nbgenclK-lastindex+1,t_VEC);
  vecp1=cgetg(lSml2+1,t_VEC);
  for (k=1; k<=lSml2; k++) vecp1[k]=lgetg(nbgenclK-lastindex+1,t_VEC);
  for (i=lastindex+1; i<=nbgenclK; i++)
  {
    p3=ginv(gmul(unmodell,(GEN)cyclicK[i]));
    p1[i-lastindex]=lmul(zideallog(nf,(GEN)listalpha[i],bid),p3);
    for (k=1; k<=lSml2; k++)
      ((GEN)vecp1[k])[i-lastindex]=
	lmul(zideallog(nf,(GEN)listalpha[i],(GEN)vsml2[k]),p3);
  }
  for (j=1; j<=lSp; j++)
  {
    p2=zideallog(nf,(GEN)vecbeta[j],bid);
    for (i=lastindex+1; i<=nbgenclK; i++)
      p2=gadd(p2,gmul(gcoeff(matexpo,i,j),(GEN)p1[i-lastindex]));
    p3=cgetg(nblig+1,t_COL); munit[j+nbvunit]=(long)p3;
    for (i=1; i<=lastbid; i++) p3[i]=p2[i];
    for (   ; i<=nblig; i++) p3[i]=coeff(matexpo,i-lastbid,j);
    for (k=1; k<=lSml2; k++)
    {
      p2=zideallog(nf,(GEN)vecbeta[j],(GEN)vsml2[k]);
      for (i=lastindex+1; i<=nbgenclK; i++)
	p2=gadd(p2,gmul(gcoeff(matexpo,i,j),gmael(vecp1,k,i-lastindex)));
      ((GEN)vecmunit2[k])[j+nbvunit]=(long)p2;
    }
    msign[j+nbvunit]=(long)zsigne(nf,(GEN)vecbeta[j],archartif,r1);
  }
  munit=gmul(unmodell,munit); vecmunit2=gmul(unmodell,vecmunit2);
  kermat=lift(ker(munit)); lker=lg(kermat)-1;
  if (!lker) err(talker,"bug2 in kummer");
  if (lSp&&(!gcmp1(gcoeff(kermat,nbcol,lker)))) err(talker,"bug3 in kummer");
  if (lker==1) { listx=cgetg(2,t_VEC); listx[1]=kermat[1]; }
  else
  {
    listx=cgetg(1,t_VEC);
    flag=1;
    yinit=cgetg(lker,t_VEC); for (i=1; i<lker; i++) yinit[i]=zero;
    y=cgetg(lker,t_VEC);
    for (i=1; i<lker; i++) { y[i]=lgeti(3); affii(gzero,(GEN)y[i]); }
    for (; flag;)
    {
      p2=cgetg(2,t_VEC);
      p3=(GEN)kermat[lker];
      for (j=1; j<lker; j++)
	p3=lift(gmul(unmodell,gadd(p3,gmul((GEN)y[j],(GEN)kermat[j]))));
      cmpt=0; for (j=1; j<=lSp; j++) if (gcmp0((GEN)p3[j+nbvunit])) cmpt++;
      if(!cmpt)
      {
	for (k=1; k<=lSml2; k++)
	{
	  p4=lift(gmul((GEN)vecmunit2[k],p3));
	  j=1; while ((j<lg(p4))&&(!signe((GEN)p4[j]))) j++;
	  if (j>=lg(p4)) break;
	}
	if (k>lSml2)
	{
	  p4=lift(gmul(msign,p3));
	  j=1; while ((j<=r1)&&gegal((GEN)p4[j],(GEN)arch[j])) j++;
	  if (j>r1) { p2[1]=(long)p3; listx=concat(listx,p2); }
	}
      }
      if (!lSp)
      {
	j=1; while ((j<lker)&&(!signe((GEN)y[j]))) j++;
	if ((j<lker)&&gcmp1((GEN)y[j]))
	{
	  p3=gsub(p3,(GEN)kermat[lker]);
	  for (k=1; k<=lSml2; k++)
	  {
	    p4=lift(gmul((GEN)vecmunit2[k],p3));
	    j=1; while ((j<lg(p4))&&(!signe((GEN)p4[j]))) j++;
	    if (j>=lg(p4)) break;
	  }
	  if (k>lSml2)
	  {
	    p4=lift(gmul(msign,p3));
	    j=1; while ((j<=r1)&&gegal((GEN)p4[j],(GEN)arch[j])) j++;
	    if (j>r1) { p2[1]=(long)p3; listx=concat(listx,p2); }
	  }
	}
      }
      i=lker;
      do
      {
	i--;
	if (!i) { flag=0; break; }
	if (i<lker-1) affii((GEN)yinit[i+1],(GEN)y[i+1]);
	gaddgsz((GEN)y[i],1,(GEN)y[i]);
      }
      while (cmpis((GEN)y[i],ell)>=0);
    }
  }
  llistx=lg(listx)-1;
  listal=cgetg(llistx+1,17);
  for (i=1; i<=llistx; i++)
  {
    p1=gun; p3=(GEN)listx[i];
    for (j=1; j<=nbvunit; j++) p1=gmul(p1,gpui((GEN)vunit[j],(GEN)p3[j],0));
    for (j=1; j<=lSp; j++) p1=gmul(p1,gpui((GEN)vecbeta[j],(GEN)p3[j+nbvunit],0));
    listal[i]=(long)p1;
  }
  
/* a ce stade, tous les alpha dans la liste listal statisfont les congruences,
   noncongruences et signatures, et ne sont pas des puissances l-iemes
   donc x^l-alpha est irreductible, sa signature est correcte ainsi que son
   discriminant relatif. Reste a determiner son groupe de normes. */
  if (DEBUGLEVEL)
  {
    fprintferr("listalpha = \n");
    outerr(listal); flusherr();
  }
  p2=cgetg(1,t_VEC);
  for (i=1; i<=llistx; i++)
  {
    p1=gsub(gpui(polx[0],gell,0),(GEN)listal[i]);
    group=rnfnormgroup(bnr,p1);
    if (all||gegal(group,subgroup)) p2=concat(p2,p1);
  }
  if (all) { tetpil=avma; return gerepile(av,tetpil,gcopy(p2)); }
  else
  {
    switch(lg(p2)-1)
    {
      case 0: err(talker,"bug 6: no equation found in kummer");
      case 1: tetpil=avma; return gerepile(av,tetpil,gcopy((GEN)p2[1]));
      default: fprintferr("equations = \n"); outerr(p2);
	err(talker,"bug 7: more than one equation found in kummer");
    }
  }
  return gzero; /* not reached */
}

